import re
from pathlib import Path
from typing import Dict, Union

from flytekit import LaunchPlan
from flytekit.configuration import SerializationSettings
from flytekit.core import constants as common_constants
from flytekit.core.base_task import PythonTask
from flytekit.core.node import Node
from flytekit.core.utils import _dnsify
from flytekit.core.workflow import WorkflowBase
from flytekit.models import common as common_models
from flytekit.models import interface as interface_models
from flytekit.models import launch_plan as launch_plan_models
from flytekit.models import task as task_models
from flytekit.models.admin import workflow as admin_workflow_models
from flytekit.models.core import identifier as identifier_model
from flytekit.models.core import workflow as workflow_model
from flytekit.models.core.workflow import TaskNodeOverrides
from typing_extensions import TypeAlias

from latch_cli.utils import urljoins

FlyteLocalEntity: TypeAlias = Union[
    PythonTask,
    Node,
    LaunchPlan,
    WorkflowBase,
]

FlyteSerializableModel: TypeAlias = Union[
    task_models.TaskSpec,
    workflow_model.Node,
    launch_plan_models.LaunchPlan,
    admin_workflow_models.WorkflowSpec,
]

EntityCache: TypeAlias = Dict[FlyteLocalEntity, FlyteSerializableModel]


def get_serializable_launch_plan(
    entity: LaunchPlan,
    settings: SerializationSettings,
    cache: EntityCache,
) -> launch_plan_models.LaunchPlan:
    if entity in cache:
        return cache[entity]

    wf_id = identifier_model.Identifier(
        resource_type=identifier_model.ResourceType.WORKFLOW,
        project=settings.project,
        domain=settings.domain,
        name=entity.workflow.name,
        version=settings.version,
    )

    lps = launch_plan_models.LaunchPlanSpec(
        workflow_id=wf_id,
        entity_metadata=launch_plan_models.LaunchPlanMetadata(
            schedule=entity.schedule,
            notifications=entity.notifications,
        ),
        default_inputs=entity.parameters,
        fixed_inputs=entity.fixed_inputs,
        labels=common_models.Labels({}),
        annotations=(
            entity.annotations
            if entity.annotations is not None
            else common_models.Annotations({})
        ),
        auth_role=None,
        raw_output_data_config=(
            entity.raw_output_data_config
            if entity.raw_output_data_config is not None
            else common_models.RawOutputDataConfig("")
        ),
        max_parallelism=entity.max_parallelism,
        security_context=entity.security_context,
    )

    lp_id = identifier_model.Identifier(
        resource_type=identifier_model.ResourceType.LAUNCH_PLAN,
        project=settings.project,
        domain=settings.domain,
        name=entity.name,
        version=settings.version,
    )
    lp_model = launch_plan_models.LaunchPlan(
        id=lp_id,
        spec=lps,
        closure=launch_plan_models.LaunchPlanClosure(
            state=None,
            expected_inputs=interface_models.ParameterMap({}),
            expected_outputs=interface_models.VariableMap({}),
        ),
    )
    cache[entity] = lp_model

    return lp_model


def get_serializable_task(
    entity: FlyteLocalEntity,
    settings: SerializationSettings,
    cache: EntityCache,
) -> task_models.TaskSpec:
    if entity in cache:
        return cache[entity]

    task_id = identifier_model.Identifier(
        identifier_model.ResourceType.TASK,
        settings.project,
        settings.domain,
        entity.name,
        settings.version,
    )

    container = entity.get_container(settings)
    pod = entity.get_k8s_pod(settings)

    tt = task_models.TaskTemplate(
        id=task_id,
        type=entity.task_type,
        metadata=entity.metadata.to_taskmetadata_model(),
        interface=entity.interface,
        custom=entity.get_custom(settings),
        container=container,
        task_type_version=entity.task_type_version,
        security_context=entity.security_context,
        config=entity.get_config(settings),
        k8s_pod=pod,
        sql=entity.get_sql(settings),
    )
    task_model = task_models.TaskSpec(template=tt)
    cache[entity] = task_model
    return task_model


class SerializationError(Exception):
    pass


def get_serializable_node(
    entity: Node,
    settings: SerializationSettings,
    cache: EntityCache,
) -> workflow_model.Node:
    if entity in cache:
        return cache[entity]

    if entity.flyte_entity is None:
        raise SerializationError(
            f"SnakemakeWorkflow Node {entity.id} has no task and cannot be serialized."
        )

    upstream_sdk_nodes = [
        get_serializable_node(n, settings, cache)
        for n in entity.upstream_nodes
        if n.id != common_constants.GLOBAL_INPUT_NODE_ID
    ]

    if isinstance(entity.flyte_entity, PythonTask):
        task_spec = get_serializable_task(entity.flyte_entity, settings, cache)
        node_model = workflow_model.Node(
            id=_dnsify(entity.id),
            metadata=entity.metadata,
            inputs=entity.bindings,
            upstream_node_ids=[n.id for n in upstream_sdk_nodes],
            output_aliases=[],
            task_node=workflow_model.TaskNode(
                reference_id=task_spec.template.id,
                overrides=TaskNodeOverrides(resources=entity._resources),
            ),
        )
        cache[entity] = node_model
        return node_model
    else:
        raise SerializationError(
            "Cannot serialize a SnakemakeWorkflow node {entity.id} containing a task"
            f" of type {entity._flyte_entity}.The task must be a PythonTask."
        )


def get_serializable_workflow(
    entity: WorkflowBase,
    settings: SerializationSettings,
    cache: EntityCache,
) -> admin_workflow_models.WorkflowSpec:
    if entity in cache:
        return cache[entity]

    upstream_node_models = [
        get_serializable_node(n, settings, cache)
        for n in entity.nodes
        if n.id != common_constants.GLOBAL_INPUT_NODE_ID
    ]

    wf_id = identifier_model.Identifier(
        resource_type=identifier_model.ResourceType.WORKFLOW,
        project=settings.project,
        domain=settings.domain,
        name=entity.name,
        version=settings.version,
    )
    wf_t = workflow_model.WorkflowTemplate(
        id=wf_id,
        metadata=entity.workflow_metadata.to_flyte_model(),
        metadata_defaults=entity.workflow_metadata_defaults.to_flyte_model(),
        interface=entity.interface,
        nodes=upstream_node_models,
        outputs=entity.output_bindings,
    )

    admin_wf = admin_workflow_models.WorkflowSpec(template=wf_t, sub_workflows=[])
    cache[entity] = admin_wf
    return admin_wf


def update_mapping(cur: Path, stem: Path, remote: str, mapping: Dict[str, str]):
    if cur.is_file():
        mapping[str(stem)] = remote
        return

    for p in cur.iterdir():
        update_mapping(p, stem / p.name, urljoins(remote, p.name), mapping)



import subprocess
from pathlib import Path

from latch.functions.messages import message
from latch.resources.tasks import small_task
from latch.types.directory import LatchOutputDir
from latch.types.file import LatchFile


@small_task
def sort_bam_task(sam: LatchFile, output_directory: LatchOutputDir) -> LatchFile:
    bam_file = Path("covid_sorted.bam").resolve()

    samtools_sort_cmd = [
        "samtools",
        "sort",
        "-o",
        str(bam_file),
        "-O",
        "bam",
        sam.local_path,
    ]

    try:
        # We use shell=True for all the benefits of pipes and other shell features.
        # When using shell=True, we pass the entire command as a single string as
        # opposed to a list since the shell will parse the string into a list
        # using its own rules.
        subprocess.run(" ".join(samtools_sort_cmd), shell=True, check=True)
    except subprocess.CalledProcessError as e:
        # will display in the messages tab of the execution graph for the sort_bam_task node
        message(
            "error",
            {"title": "Samtools Sort Failed", "body": f"Error: {str(e)}"},
        )
        raise e

    # intended output path of the file in Latch console, constructed from
    # the user provided output directory
    output_location = f"{output_directory.remote_directory}/covid_sorted.bam"

    return LatchFile(str(bam_file), output_location)



import os
import sys
import termios
import tty
from typing import Any, Callable, Generic, List, Optional, Tuple, TypeVar

from typing_extensions import TypedDict

from latch_cli.click_utils import AnsiCodes


def buffered_print() -> Tuple[Callable, Callable]:
    buffer = []

    def __print(*args):
        for arg in args:
            buffer.append(arg)

    def __show():
        nonlocal buffer
        print("".join(buffer), flush=True, end="")
        buffer = []

    return __print, __show


# Allows for exactly one print per render, removing any weird flashing
# behavior and also speeding things up considerably
_print, _show = buffered_print()


def clear(k: int):
    """
    Clear `k` lines below the cursor, returning the cursor to its original position
    """
    _print(f"\x1b[2K\x1b[1E" * (k) + f"\x1b[{k}F")


def draw_box(
    ul_corner_pos: Tuple[int, int],
    height: int,
    width: int,
    color: Optional[str] = None,
):
    if height <= 0 or width <= 0:
        return
    move_cursor(ul_corner_pos)
    draw_horizontal_line(width, make_corner=True, color=color)
    draw_vertical_line(height, make_corner=True, color=color)
    draw_horizontal_line(width, left=True, make_corner=True, color=color)
    draw_vertical_line(height, up=True, make_corner=True, color=color)


def clear_screen():
    _print("\x1b[2J")


def remove_cursor():
    _print("\x1b[?25l")


def reveal_cursor():
    _print("\x1b[?25h")


def move_cursor(pos: Tuple[int, int]):
    """
    Move the cursor to a given (x, y) coordinate
    """
    x, y = pos
    if x < 0 or y < 0:
        return
    _print(f"\x1b[{y};{x}H")


def move_cursor_up(n: int):
    if n <= 0:
        return
    _print(f"\x1b[{n}A")


def line_up(n: int):
    """Moves to the start of the destination line"""
    if n <= 0:
        return
    _print(f"\x1b[{n}F")


def move_cursor_down(n: int):
    if n <= 0:
        return
    _print(f"\x1b[{n}B")


def line_down(n: int):
    """Moves to the start of the destination line"""
    if n <= 0:
        return
    _print(f"\x1b[{n}E")


def move_cursor_right(n: int):
    if n <= 0:
        return
    _print(f"\x1b[{n}C")


def move_cursor_left(n: int):
    if n <= 0:
        return
    _print(f"\x1b[{n}D")


def current_cursor_position() -> Tuple[int, int]:
    res = b""
    sys.stdout.write("\x1b[6n")
    sys.stdout.flush()
    while not res.endswith(b"R"):
        res += sys.stdin.buffer.read(1)
    y, x = res.strip(b"\x1b[R").split(b";")
    return int(x), int(y)


def draw_vertical_line(
    height: int,
    up: bool = False,
    make_corner: bool = False,
    color: Optional[str] = None,
):
    """
    Draws a vertical line with given `height`, going upwards if `up` is True
    and downwards otherwise.
    """

    if height <= 0:
        return

    if color is not None:
        _print(color)
    sep = "\x1b[1A" if up else "\x1b[1B"
    for i in range(height):
        if i == 0 and make_corner:
            corner = "\u2514" if up else "\u2510"
            _print(f"{corner}\x1b[1D{sep}")
        else:
            _print(f"\u2502\x1b[1D{sep}")
    if color is not None:
        _print("\x1b[0m")


def draw_horizontal_line(
    width: int,
    left: bool = False,
    make_corner: bool = False,
    color: Optional[str] = None,
):
    """
    Draws a horizontal line with given `width`, going to the left if `left` is True
    and to the right otherwise.
    """

    if width <= 0:
        return

    if color is not None:
        _print(color)
    sep = "\x1b[2D" if left else ""
    for i in range(width):
        if i == 0 and make_corner:
            corner = "\u2518" if left else "\u250c"
            _print(f"{corner}{sep}")
        else:
            _print(f"\u2500{sep}")
    if color is not None:
        _print("\x1b[0m")


def read_next_byte() -> bytes:
    b = sys.stdin.buffer.read(1)
    if b in (
        b"\x03",  # CTRL C
        b"\x04",  # CTRL D
        b"q",
        b"Q",
    ):
        raise KeyboardInterrupt
    return b


def read_bytes(num_bytes: int) -> bytes:
    if num_bytes < 0:
        raise ValueError(f"cannot read {num_bytes} bytes")
    result = b""
    for _ in range(num_bytes):
        result += read_next_byte()
    return result


T = TypeVar("T")


class SelectOption(TypedDict, Generic[T]):
    display_name: str
    value: T


def select_tui(
    title: str, options: List[SelectOption[T]], clear_terminal: bool = True
) -> Optional[T]:
    """
    Renders a terminal UI that allows users to select one of the options
    listed in `options`

    Args:
        title: The title of the selection window.
        options: A list of names for each of the options.
        clear_terminal: Whether or not to clear the entire terminal window
            before displaying - default False
    """

    if len(options) == 0:
        raise ValueError("No options given")

    def render(
        curr_selected: int,
        start_index: int = 0,
        max_per_page: int = 10,
        indent: str = "    ",
    ) -> int:
        if curr_selected < 0 or curr_selected >= len(options):
            curr_selected = 0

        _print(title)
        line_down(2)

        num_lines_rendered = 4  # 4 "extra" lines for header + footer

        for i in range(start_index, start_index + max_per_page):
            if i >= len(options):
                break
            name = options[i]["display_name"]
            if i == curr_selected:
                color = AnsiCodes.color
                bold = AnsiCodes.bold
                reset = AnsiCodes.full_reset

                prefix = indent[:-2] + "> "

                _print(f"{color}{bold}{prefix}{name}{reset}\x1b[1E")
            else:
                _print(f"{indent}{name}\x1b[1E")
            num_lines_rendered += 1

        line_down(1)

        control_str = "[ARROW-KEYS] Navigate\t[ENTER] Select\t[Q] Quit"
        _print(control_str)
        line_up(num_lines_rendered - 1)

        _show()

        return num_lines_rendered

    old_settings = termios.tcgetattr(sys.stdin.fileno())
    tty.setraw(sys.stdin.fileno())

    curr_selected = 0
    start_index = 0
    _, term_height = os.get_terminal_size()
    remove_cursor()

    max_per_page = min(len(options), term_height - 4)

    if clear_terminal:
        clear_screen()
        move_cursor((0, 0))
    else:
        print("\n" * (max_per_page + 3))
        move_cursor_up(max_per_page + 4)

    num_lines_rendered = render(
        curr_selected,
        start_index=start_index,
        max_per_page=max_per_page,
    )

    try:
        while True:
            b = read_bytes(1)
            if b == b"\r":
                return options[curr_selected]["value"]
            elif b == b"\x1b":
                b = read_bytes(2)
                if b == b"[A":  # Up Arrow
                    curr_selected = max(curr_selected - 1, 0)
                    if (
                        curr_selected - start_index < max_per_page // 2
                        and start_index > 0
                    ):
                        start_index -= 1
                elif b == b"[B":  # Down Arrow
                    curr_selected = min(curr_selected + 1, len(options) - 1)
                    if (
                        curr_selected - start_index > max_per_page // 2
                        and start_index < len(options) - max_per_page
                    ):
                        start_index += 1
                else:
                    continue
            clear(num_lines_rendered)
            num_lines_rendered = render(
                curr_selected,
                start_index=start_index,
                max_per_page=max_per_page,
            )
    except KeyboardInterrupt:
        ...
    finally:
        clear(num_lines_rendered)
        reveal_cursor()
        _show()
        termios.tcsetattr(sys.stdin.fileno(), termios.TCSANOW, old_settings)



import base64
import contextlib
import importlib.machinery as im
import importlib.util as iu
import io
import os
import sys
import typing
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    TypedDict,
    Union,
)

import boto3
import docker
import requests
from latch_sdk_config.latch import config

from ...utils import current_workspace

if TYPE_CHECKING:
    from ...centromere.ctx import _CentromereCtx
else:
    _CentromereCtx = ""


# todo(maximsmol): only login if the credentials are expired
def _docker_login(ctx: _CentromereCtx):
    assert ctx.dkr_client is not None

    headers = {"Authorization": f"Bearer {ctx.token}"}
    data = {"pkg_name": ctx.image, "ws_account_id": current_workspace()}
    response = requests.post(ctx.latch_image_api_url, headers=headers, json=data)

    try:
        response = response.json()
        access_key = response["tmp_access_key"]
        secret_key = response["tmp_secret_key"]
        session_token = response["tmp_session_token"]
    except KeyError as err:
        raise ValueError(f"malformed response on image upload: {response}") from err

    try:
        client = boto3.session.Session(
            aws_access_key_id=access_key,
            aws_secret_access_key=secret_key,
            aws_session_token=session_token,
            region_name="us-west-2",
        ).client("ecr")
        token = client.get_authorization_token()["authorizationData"][0][
            "authorizationToken"
        ]
    except Exception as err:
        raise ValueError(
            f"unable to retreive an ecr login token for user {ctx.account_id}"
        ) from err

    auth = ctx.dkr_client._auth_configs
    store_name = auth.get_credential_store(ctx.dkr_repo)
    if store_name is not None:
        store = auth._get_store_instance(store_name)
        try:
            store.erase(ctx.dkr_repo)
        # To handle: "Credentials store docker-credential-osxkeychain exited
        # with "The specified item could not be found in the keychain.""
        except docker.credentials.errors.StoreError:
            pass

    user, password = base64.b64decode(token).decode("utf-8").split(":")
    res = ctx.dkr_client.login(
        username=user, password=password, registry=ctx.dkr_repo, reauth=True
    )
    assert res["Status"] == "Login Succeeded"


class DockerBuildLogItem(TypedDict):
    message: Optional[str]
    error: Optional[str]
    stream: Optional[str]


def build_image(
    ctx: _CentromereCtx,
    image_name: str,
    context_path: Path,
    dockerfile: Optional[Path] = None,
) -> Iterable[DockerBuildLogItem]:
    assert ctx.dkr_client is not None

    _docker_login(ctx)
    build_logs = ctx.dkr_client.build(
        path=str(context_path),
        dockerfile=str(dockerfile) if dockerfile is not None else None,
        buildargs={"tag": f"{ctx.dkr_repo}/{image_name}"},
        tag=f"{ctx.dkr_repo}/{image_name}",
        decode=True,
    )

    return build_logs


def upload_image(ctx: _CentromereCtx, image_name: str) -> List[str]:
    assert ctx.dkr_client is not None
    return ctx.dkr_client.push(
        repository=f"{ctx.dkr_repo}/{image_name}",
        stream=True,
        decode=True,
    )


def serialize_pkg_in_container(
    ctx: _CentromereCtx, image_name: str, serialize_dir: str
) -> Tuple[List[str], str]:
    assert ctx.dkr_client is not None

    _serialize_cmd = ["make", "serialize"]
    container = ctx.dkr_client.create_container(
        f"{ctx.dkr_repo}/{image_name}",
        command=_serialize_cmd,
        volumes=[serialize_dir],
        environment={"LATCH_DKR_REPO": ctx.dkr_repo, "LATCH_VERSION": ctx.version},
        host_config=ctx.dkr_client.create_host_config(
            binds={
                serialize_dir: {
                    "bind": "/tmp/output",
                    "mode": "rw",
                },
            }
        ),
    )
    container_id = typing.cast(str, container.get("Id"))
    ctx.dkr_client.start(container_id)
    logs = typing.cast(Iterable[bytes], ctx.dkr_client.logs(container_id, stream=True))

    return [x.decode("utf-8") for x in logs], container_id


def register_serialized_pkg(
    files: List[Path],
    token: Optional[str],
    version: str,
    workspace_id: str,
    latch_register_url: str = config.api.workflow.register,
) -> object:
    if token is None:
        token = os.environ.get("FLYTE_INTERNAL_EXECUTION_ID", "")
        if token != "":
            headers = {"Authorization": f"Latch-Execution-Token {token}"}
        else:
            raise OSError(
                "The environment variable FLYTE_INTERNAL_EXECUTION_ID does not exist"
            )
    else:
        headers = {"Authorization": f"Bearer {token}"}

    serialize_files: Dict[str, Union[bytes, io.BufferedReader]] = {
        "version": version.encode("utf-8"),
        ".latch_ws": workspace_id.encode("utf-8"),
    }
    with contextlib.ExitStack() as stack:
        for file in files:
            fh = open(file, "rb")
            stack.enter_context(fh)

            serialize_files[fh.name] = fh

        response = requests.post(
            latch_register_url,
            headers=headers,
            files=serialize_files,
        )
        response.raise_for_status()
        return response.json()


def import_module_by_path(x: Path, *, module_name: str = "latch_metadata"):
    spec = iu.spec_from_file_location(module_name, x)
    assert spec is not None
    assert spec.loader is not None

    module = iu.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)

    return module



from textwrap import dedent

import click
import gql
from gql.transport.exceptions import TransportQueryError
from latch_sdk_gql.execute import execute

from latch.ldata._transfer.node import get_node_data as _get_node_data
from latch.ldata.type import LDataNodeType
from latch_cli.services.cp.glob import expand_pattern
from latch_cli.utils.path import get_name_from_path, get_path_error, is_remote_path


def move(
    src: str,
    dest: str,
    *,
    no_glob: bool = False,
):
    if not is_remote_path(src) or not is_remote_path(dest):
        click.secho(
            f"`latch mv` cannot be used for local file operations. Please make sure"
            f" all of your input paths are remote (beginning with `latch://`)",
            fg="red",
        )
        raise click.exceptions.Exit(1)

    if no_glob:
        srcs = [src]
    else:
        srcs = expand_pattern(src)

    if len(srcs) == 0:
        click.echo(f"Could not find any files that match pattern {src}. Exiting.")
        raise click.exceptions.Exit(0)

    try:
        node_data = _get_node_data(*srcs, dest, allow_resolve_to_parent=True)
    except Exception as e:
        click.echo(str(e))
        raise click.exceptions.Exit(1) from e

    dest_data = node_data.data[dest]
    acc_id = node_data.acc_id

    multi_src = len(srcs) > 1

    if not (dest_data.exists() or dest_data.is_direct_parent()):
        click.echo(get_path_error(dest, "not found", acc_id))
        raise click.exceptions.Exit(1)
    elif multi_src and not dest_data.exists():
        click.secho(
            f"Remote destination {dest} does not exist. Cannot move multiple source"
            " files to a destination that does not exist.",
            fg="red",
        )
        raise click.exceptions.Exit(1)
    elif multi_src and dest_data.type in {LDataNodeType.obj, LDataNodeType.link}:
        click.secho(
            f"Remote destination {dest} is not a directory. Cannot move multiple source"
            " files to a destination that is not a directory.",
            fg="red",
        )
        raise click.exceptions.Exit(1)

    path_by_id = {v.id: k for k, v in node_data.data.items()}

    for s in srcs:
        src_data = node_data.data[s]

        if not src_data.exists():
            click.echo(get_path_error(s, "not found", acc_id))
            raise click.exceptions.Exit(1)

        new_name = None
        if dest_data.is_direct_parent():
            new_name = get_name_from_path(dest)
        elif dest_data.type in {LDataNodeType.obj, LDataNodeType.link}:
            click.echo(get_path_error(dest, "object already exists at path.", acc_id))
            raise click.exceptions.Exit(1)

        try:
            execute(
                gql.gql("""
                    mutation Move(
                        $argNode: BigInt!
                        $argDestParent: BigInt!
                        $argNewName: String
                    ) {
                        ldataMove(
                            input: {
                                argNode: $argNode
                                argDestParent: $argDestParent
                                argNewName: $argNewName
                            }
                        ) {
                            clientMutationId
                        }
                    }
                """),
                {
                    "argNode": src_data.id,
                    "argDestParent": dest_data.id,
                    "argNewName": new_name,
                },
            )
        except TransportQueryError as e:
            if e.errors is None or len(e.errors) == 0:
                raise e

            msg: str = e.errors[0]["message"]

            if msg.startswith("Permission denied on node"):
                node_id = msg.rsplit(" ", 1)[1]
                path = path_by_id[node_id]

                click.echo(get_path_error(path, "permission denied.", acc_id))
                raise click.exceptions.Exit(1) from e
            elif msg == "Refusing to make node its own parent":
                click.echo(get_path_error(dest, f"is a parent of {s}.", acc_id))
                raise click.exceptions.Exit(1) from e
            elif msg == "Refusing to parent node to an object node":
                click.echo(get_path_error(dest, f"object exists at path.", acc_id))
                raise click.exceptions.Exit(1) from e
            elif msg == "Refusing to move a share link (or into a share link)":
                if src_data.type is LDataNodeType.link:
                    path = s
                else:
                    path = dest

                click.echo(get_path_error(path, f"is a share link.", acc_id))
                raise click.exceptions.Exit(1) from e
            elif msg.startswith("Refusing to move account root"):
                click.echo(get_path_error(s, "is an account root.", acc_id))
                raise click.exceptions.Exit(1) from e
            elif msg.startswith("Refusing to move removed node"):
                click.echo(get_path_error(s, "not found.", acc_id))
                raise click.exceptions.Exit(1) from e
            elif msg.startswith("Refusing to move already moved node"):
                click.echo(
                    get_path_error(
                        s,
                        "copy in progress. Please wait until the node has finished"
                        " copying before moving.",
                        acc_id,
                    )
                )
                raise click.exceptions.Exit(1) from e
            elif msg == "Conflicting object in destination":
                click.echo(get_path_error(dest, "object exists at path.", acc_id))
                raise click.exceptions.Exit(1) from e
            elif msg.startswith("Refusing to do noop move"):
                click.echo(get_path_error(dest, "cannot move node to itself.", acc_id))
                raise click.exceptions.Exit(1) from e
            else:
                raise e

    if len(srcs) == 1:
        src_str = f'{click.style("Source: ", fg="blue")}{srcs[0]}'
    else:
        src_str = "\n".join(
            [click.style("Sources: ", fg="blue"), *[f"    {s}" for s in srcs]]
        )

    click.echo(dedent(f"""
        {click.style("Move Succeeded.", fg="green")}

        __srcs__
        {click.style("Destination: ", fg="blue")}{(dest)}
        """).replace("__srcs__", src_str).strip())



from dataclasses import dataclass
from typing import Dict, List, TypedDict

import graphql.language as l
from latch_sdk_gql.utils import _name_node, _parse_selection
from typing_extensions import TypeAlias

from latch.ldata.type import LatchPathError, LDataNodeType
from latch_cli.utils.path import normalize_path

from .utils import query_with_retry

AccId: TypeAlias = int


class LDataObjectMeta(TypedDict):
    contentSize: str
    contentType: str


class FinalLinkTargetPayload(TypedDict):
    id: str
    type: str
    name: str
    removed: bool
    ldataObjectMeta: LDataObjectMeta


class LdataNodePayload(TypedDict):
    finalLinkTarget: FinalLinkTargetPayload


class LdataResolvePathToNodePayload(TypedDict):
    path: str
    ldataNode: LdataNodePayload


class AccountInfoCurrentPayload(TypedDict):
    id: str


@dataclass(frozen=True)
class NodeData:
    id: str
    name: str
    type: LDataNodeType
    remaining: str

    def is_direct_parent(self) -> bool:
        return self.remaining is not None and "/" not in self.remaining

    def exists(self) -> bool:
        return self.remaining is None or self.remaining == ""


@dataclass(frozen=True)
class GetNodeDataResult:
    acc_id: str
    data: Dict[str, NodeData]


def get_node_data(
    *remote_paths: str, allow_resolve_to_parent: bool = False
) -> GetNodeDataResult:
    normalized: Dict[str, str] = {}

    acc_sel = _parse_selection("""
        accountInfoCurrent {
            id
        }
    """)
    assert isinstance(acc_sel, l.FieldNode)

    sels: List[l.FieldNode] = [acc_sel]

    for i, remote_path in enumerate(remote_paths):
        normalized[remote_path] = normalize_path(remote_path)

        sel = _parse_selection("""
            ldataResolvePathToNode(path: {}) {
                path
                ldataNode {
                    finalLinkTarget {
                        id
                        name
                        type
                    }
                }
            }
        """)
        assert isinstance(sel, l.FieldNode)

        val = l.StringValueNode()
        val.value = normalized[remote_path]

        args = l.ArgumentNode()
        args.name = _name_node("path")
        args.value = val

        sel.alias = _name_node(f"q{i}")
        sel.arguments = (args,)

        sels.append(sel)

    sel_set = l.SelectionSetNode()
    sel_set.selections = tuple(sels)

    doc = l.parse("""
        query GetNodeType {
            placeholder
        }
        """)

    assert len(doc.definitions) == 1
    query = doc.definitions[0]

    assert isinstance(query, l.OperationDefinitionNode)
    query.selection_set = sel_set

    res = query_with_retry(doc)

    acc_info: AccountInfoCurrentPayload = res["accountInfoCurrent"]
    acc_id = acc_info["id"]

    ret: Dict[str, NodeData] = {}
    for i, remote_path in enumerate(remote_paths):
        node: LdataResolvePathToNodePayload = res[f"q{i}"]

        try:
            remaining = node["path"]
            if (
                remaining is not None and remaining != ""
            ) and not allow_resolve_to_parent:
                raise LatchPathError(
                    f"no such Latch file or directory", remote_path, acc_id
                )

            final_link_target = node["ldataNode"]["finalLinkTarget"]
            ret[remote_path] = NodeData(
                id=final_link_target["id"],
                name=final_link_target["name"],
                type=LDataNodeType(final_link_target["type"].lower()),
                remaining=remaining,
            )
        except (TypeError, ValueError):
            raise LatchPathError(
                f"no such Latch file or directory", remote_path, acc_id
            )

    return GetNodeDataResult(acc_id, ret)



from flytekit.core.launch_plan import reference_launch_plan

from latch_cli.utils import current_workspace


def workflow_reference(
    name: str,
    version: str,
):
    return reference_launch_plan(
        project=current_workspace(),
        domain="development",
        name=name,
        version=version,
    )



from contextlib import contextmanager
from enum import Enum
from multiprocessing import BoundedSemaphore
from typing import Dict, List, Optional

import tqdm


def get_progress_bar():
    return tqdm.tqdm(
        total=0,
        leave=False,
        smoothing=0,
        unit="B",
        unit_scale=True,
    )


class Progress(Enum):
    none = "none"
    total = "total"
    tasks = "tasks"


class ProgressBars:
    def __init__(
        self,
        num_task_bars: int,
        *,
        show_total_progress: bool = True,
        verbose: bool = False,
    ):
        if show_total_progress:
            self.total_bar = get_progress_bar()
            self.total_bar.desc = "Copying Files"
            self.total_bar.colour = "green"
            self.total_bar.unit = ""
        else:
            self.total_bar = None

        self.total_sema = BoundedSemaphore(1)

        self.verbose = verbose

        self.task_bars: List[tqdm.tqdm] = [
            get_progress_bar() for _ in range(num_task_bars)
        ]
        self.free_indices = {i for i in range(num_task_bars)}
        self.task_bar_sema = BoundedSemaphore(num_task_bars)

        self.usage: Dict[str, int] = {}

    def num_bars(self) -> int:
        return len(self.task_bars)

    def write(self, msg: str):
        if not self.verbose:
            return

        tqdm.tqdm.write(msg)

    def get_free_task_bar_index(self) -> Optional[int]:
        if len(self.task_bars) == 0:
            return None

        self.task_bar_sema.acquire(block=True)
        return self.free_indices.pop()

    def return_task_bar(self, index: Optional[int]):
        if index is None:
            return

        self.reset(index)
        self.free_indices.add(index)
        self.task_bar_sema.release()

    def set_usage(self, key: str, amount: int):
        self.usage[key] = amount

    def dec_usage(self, key: str):
        if key not in self.usage:
            return

        amount = self.usage[key]
        amount -= 1
        self.usage[key] = amount
        return amount

    def set_total(self, total: int, desc: Optional[str] = None):
        if self.total_bar is None:
            return

        self.total_bar.total = total
        if desc is not None:
            self.total_bar.desc = desc

    def update_total_progress(self, amount: int):
        if self.total_bar is None:
            return

        with self.total_sema:
            self.total_bar.update(amount)
            self.total_bar.refresh()

    def set(self, index: Optional[int], total: int, desc: str):
        if index is None:
            return

        self.task_bars[index].total = total
        self.task_bars[index].desc = desc
        self.task_bars[index].refresh()

    def update(self, index: Optional[int], amount: int):
        if index is None:
            return

        self.task_bars[index].update(amount)
        self.task_bars[index].refresh()

    def reset(self, index: Optional[int]):
        if index is None:
            return

        self.task_bars[index].reset()
        self.task_bars[index].total = 0
        self.task_bars[index].desc = ""
        self.task_bars[index].refresh()

    def close(self):
        if self.total_bar is not None:
            self.total_bar.close()

        for bar in self.task_bars:
            bar.close()

        # move cursor back to beginning of line
        print("\r", end="")


@contextmanager
def get_free_index(progress_bars: ProgressBars):
    try:
        pbar_index = progress_bars.get_free_task_bar_index()
        yield pbar_index
    finally:
        progress_bars.return_task_bar(pbar_index)



import os
import secrets
import string
from pathlib import Path

import pytest


@pytest.fixture(scope="session")
def test_account_jwt():
    tmp_token = os.environ["TEST_TOKEN"]
    token_dir = Path.home().joinpath(".latch")
    token_dir.mkdir(exist_ok=True)
    token_file = token_dir.joinpath("token")
    with open(token_file, "w") as f:
        f.write(tmp_token)

    return tmp_token


@pytest.fixture(scope="session")
def project_name():
    alphabet = string.ascii_letters
    return "".join(secrets.choice(alphabet) for i in range(8)).lower()



from contextlib import contextmanager
from dataclasses import dataclass, field

try:
    from functools import cache
except ImportError:
    from functools import lru_cache as cache

from typing import Iterator, List, Literal, Optional, TypedDict, Union, overload

import gql
import graphql.language as l
from latch_sdk_config.user import user_config
from latch_sdk_gql.execute import execute
from latch_sdk_gql.utils import _GqlJsonValue, _json_value, _name_node, _parse_selection
from typing_extensions import Self, TypeAlias

from latch.registry.project import Project
from latch.registry.table import Table


class _CatalogExperiment(TypedDict):
    id: str
    displayName: str


class _CatalogExperimentConnection(TypedDict):
    nodes: List[_CatalogExperiment]


class _CatalogProjectNode(TypedDict):
    id: str
    displayName: str

    catalogExperimentsByProjectId: _CatalogExperimentConnection


class _CatalogProjectConnection(TypedDict):
    nodes: List[_CatalogProjectNode]


class _Account(TypedDict):
    catalogProjectsByOwnerId: _CatalogProjectConnection


@dataclass
class _Cache:
    catalog_projects: Optional[List[Project]] = None


@dataclass(frozen=True)
class Account:
    """User or team workspace. Can be used to fetch related resources.

    :meth:`current` is the typical way of getting an :class:`Account`.

    If the current request signer (CLI user or execution context)
    lacks permissions to fetch some information, the corresponding operations
    will act as if the information does not exist. Update operations will usually
    produce errors.
    """

    _cache: _Cache = field(
        default_factory=lambda: _Cache(),
        init=False,
        repr=False,
        hash=False,
        compare=False,
    )

    id: str
    """Unique identifier."""

    @classmethod
    @cache
    def current(cls) -> Self:
        """Get current account.

        In an execution context, this is the workspace in which the execution
        was run.

        In the CLI context (when running `latch` commands) this is the
        current setting of `latch workspace`, which defaults to the user's personal
        workspace.

        Returns:
            Current account.
        """
        if user_config.workspace_id != "":
            account_id = user_config.workspace_id
        else:
            account_id = execute(
                document=gql.gql("""
                    query accountInfoQuery {
                        accountInfoCurrent {
                            id
                        }
                    }
                    """),
            )["accountInfoCurrent"]["id"]

        return cls(id=account_id)

    def load(self) -> None:
        """(Re-)populate this account instance's cache.

        Future calls to most getters will return immediately without making a network request.

        Always makes a network request.
        """
        data: _Account = execute(
            gql.gql("""
            query AccountQuery($ownerId: BigInt!) {
                accountInfo(id: $ownerId) {
                    catalogProjectsByOwnerId(
                        condition: {
                            removed: false
                        }
                    ) {
                        nodes {
                            id
                            displayName

                            catalogExperimentsByProjectId(condition: {removed: false}) {
                                nodes {
                                    id
                                    displayName
                                }
                            }
                        }
                    }
                }
            }
            """),
            {"ownerId": self.id},
        )["accountInfo"]
        # todo(maximsmol): deal with nonexistent accounts

        self._cache.catalog_projects = []
        for x in data["catalogProjectsByOwnerId"]["nodes"]:
            cur = Project(x["id"])
            self._cache.catalog_projects.append(cur)

            cur._cache.display_name = x["displayName"]

            cur._cache.tables = []
            for t in x["catalogExperimentsByProjectId"]["nodes"]:
                table = Table(t["id"])
                cur._cache.tables.append(table)

                table._cache.display_name = t["displayName"]

    # list_registry_projects

    @overload
    def list_registry_projects(
        self, *, load_if_missing: Literal[True] = True
    ) -> List[Project]: ...

    @overload
    def list_registry_projects(
        self, *, load_if_missing: bool
    ) -> Optional[List[Project]]: ...

    def list_registry_projects(
        self, *, load_if_missing: bool = True
    ) -> Optional[List[Project]]:
        """List Registry projects owned by this workspace.

        Args:
            load_if_missing:
                If true, :meth:`load` the project list if not in cache.
                If false, return `None` if not in cache.

        Returns:
            Projects owned by this workspace.
        """
        if self._cache.catalog_projects is None and load_if_missing:
            self.load()

        return self._cache.catalog_projects

    @contextmanager
    def update(self, *, reload_on_commit: bool = True) -> Iterator["AccountUpdate"]:
        """Start an update transaction.

        The transaction will commit when the context manager closes unless an error occurs.

        No changes will occur until the transaction commits.

        The transaction can be cancelled by running :meth:`AccountUpdate.clear`
        before closing the context manager.

        Args:
            reload_on_commit:
                If true, :meth:`load` this account after the transaction commits.

        Returns:
            Context manager for the new transaction.
        """

        upd = AccountUpdate(self)
        yield upd
        upd.commit()

        if reload_on_commit:
            self.load()

    def __repr__(self):
        return f"Account(id={self.id})"

    def __str__(self):
        return repr(self)


@dataclass(frozen=True)
class _AccountRegistryProjectsUpsertData:
    display_name: str


@dataclass(frozen=True)
class _AccountRegistryProjectsDeleteData:
    id: str


_AccountMutationsData: TypeAlias = Union[
    _AccountRegistryProjectsUpsertData, _AccountRegistryProjectsDeleteData
]


@dataclass(frozen=True)
class AccountUpdate:
    _mutations: List[_AccountMutationsData] = field(
        default_factory=list,
        init=False,
        repr=False,
        hash=False,
        compare=False,
    )

    account: Account

    # upsert registry project

    def upsert_registry_project(self, display_name: str):
        """Upsert a registry project.

        Not idempotent. Two calls with the same args will create two projects.

        Args:
            display_name: Display name of the new project.
        """
        self._mutations.append(_AccountRegistryProjectsUpsertData(display_name))

    def _add_registry_projects_upsert_selection(
        self,
        upserts: List[_AccountRegistryProjectsUpsertData],
        mutations: List[l.SelectionNode],
    ):
        if len(upserts) == 0:
            return

        display_names: _GqlJsonValue = [x.display_name for x in upserts]

        res = _parse_selection("""
            catalogMultiCreateProjects(input: {}) {
                clientMutationId
            }
            """)
        assert isinstance(res, l.FieldNode)

        args = l.ArgumentNode()
        args.name = _name_node("input")
        args.value = _json_value(
            {
                "argOwnerId": self.account.id,
                "argDisplayNames": display_names,
            }
        )

        res.alias = _name_node(f"upd{len(mutations)}")
        res.arguments = tuple([args])

        mutations.append(res)

    # delete registry project

    def delete_registry_project(self, id: str):
        """Delete a registry project.

        Args:
            id: The ID of the target project.
        """
        self._mutations.append(_AccountRegistryProjectsDeleteData(id))

    def _add_registry_projects_delete_selection(
        self,
        deletes: List[_AccountRegistryProjectsDeleteData],
        mutations: List[l.SelectionNode],
    ):
        if len(deletes) == 0:
            return

        ids: _GqlJsonValue = [x.id for x in deletes]

        res = _parse_selection("""
            catalogMultiDeleteProjects(input: {}) {
                clientMutationId
            }
            """)
        assert isinstance(res, l.FieldNode)

        args = l.ArgumentNode()
        args.name = _name_node("input")
        args.value = _json_value(
            {
                "argIds": ids,
            }
        )

        res.alias = _name_node(f"upd{len(mutations)}")
        res.arguments = tuple([args])

        mutations.append(res)

    # transaction

    def commit(self) -> None:
        """Commit this account update transaction.

        May be called multiple times.

        All pending updates are committed with one network request.

        Atomic. The entire transaction either commits or fails with an exception.
        """
        mutations: List[l.SelectionNode] = []

        if len(self._mutations) == 0:
            return

        def _add_mutations_selection(cur):
            if isinstance(cur[0], _AccountRegistryProjectsUpsertData):
                self._add_registry_projects_upsert_selection(cur, mutations)
            if isinstance(cur[0], _AccountRegistryProjectsDeleteData):
                self._add_registry_projects_delete_selection(cur, mutations)

        cur = [self._mutations[0]]
        for mut in self._mutations[1:]:
            if isinstance(mut, type(cur[0])):
                cur.append(mut)
                continue

            _add_mutations_selection(cur)
            cur = [mut]

        _add_mutations_selection(cur)

        sel_set = l.SelectionSetNode()
        sel_set.selections = tuple(mutations)

        doc = l.parse("""
            mutation AccountUpdate {
                placeholder
            }
        """)

        assert len(doc.definitions) == 1
        mut = doc.definitions[0]

        assert isinstance(mut, l.OperationDefinitionNode)
        mut.selection_set = sel_set

        execute(doc)

        self.clear()

    def clear(self):
        """Remove pending updates.

        May be called to cancel any pending updates that have not been committed.
        """
        self._mutations.clear()



import atexit
import os
import re
import shutil
from dataclasses import dataclass, field
from pathlib import Path
from typing import Iterator, Optional, Type

import gql
from flytekit import (
    Blob,
    BlobMetadata,
    BlobType,
    FlyteContext,
    Literal,
    LiteralType,
    Scalar,
)
from flytekit.extend import TypeEngine, TypeTransformer
from typing_extensions import Self

from latch.ldata.type import LatchPathError, LDataNodeType
from latch_cli.utils import urljoins

from ._transfer.download import download as _download
from ._transfer.node import get_node_data as _get_node_data
from ._transfer.progress import Progress as _Progress
from ._transfer.remote_copy import remote_copy as _remote_copy
from ._transfer.upload import upload as _upload
from ._transfer.utils import query_with_retry

node_id_regex = re.compile(r"^latch://(?P<id>[0-9]+)\.node$")


_dir_types = {
    LDataNodeType.dir,
    LDataNodeType.account_root,
    LDataNodeType.mount,
}

_download_idx = 0


@dataclass
class _Cache:
    path: Optional[str] = None
    node_id: Optional[str] = None
    name: Optional[str] = None
    type: Optional[LDataNodeType] = None
    size: Optional[int] = None
    content_type: Optional[str] = None


@dataclass(frozen=True)
class LPath:
    """Latch Path.

    Represents a remote file/directory path hosted on Latch. Can be used to
    interact with files and directories in Latch.

    Attributes:
    path: The Latch path. Must start with "latch://".
    """

    _cache: _Cache = field(
        default_factory=_Cache,
        init=False,
        repr=False,
        hash=False,
        compare=False,
    )

    path: str

    def __post_init__(self):
        if isinstance(self.path, LPath):
            raise ValueError("LPath cannot be initialized with another LPath")
        if not self.path.startswith("latch://"):
            raise ValueError(f"invalid LPath: {self.path} is not a Latch path")

    def fetch_metadata(self) -> None:
        """(Re-)populate this LPath's instance's cache.

        Future calls to most getters will return immediately without making a network request.

        Always makes a network request.
        """
        data = query_with_retry(
            gql.gql("""
            query GetNodeData($path: String!) {
                ldataResolvePathToNode(path: $path) {
                    path
                    ldataNode {
                        finalLinkTarget {
                            id
                            name
                            type
                            ldataObjectMeta {
                                contentSize
                                contentType
                            }
                        }
                    }
                }
            }"""),
            {"path": self.path},
        )["ldataResolvePathToNode"]

        if (
            data is None
            or data["ldataNode"] is None
            or (data["path"] is not None and data["path"] != "")
        ):
            raise LatchPathError(f"no such Latch file or directory", self.path)

        self._cache.path = self.path

        final_link_target = data["ldataNode"]["finalLinkTarget"]
        self._cache.node_id = final_link_target["id"]
        self._cache.name = final_link_target["name"]
        self._cache.type = LDataNodeType(final_link_target["type"].lower())

        meta = final_link_target["ldataObjectMeta"]
        if meta is not None:
            self._cache.size = (
                None if meta["contentSize"] is None else int(meta["contentSize"])
            )
            self._cache.content_type = meta["contentType"]

    def _clear_cache(self):
        self._cache.path = None
        self._cache.node_id = None
        self._cache.name = None
        self._cache.type = None
        self._cache.size = None
        self._cache.content_type = None

    def node_id(self, *, load_if_missing: bool = True) -> Optional[str]:
        match = node_id_regex.match(self.path)
        if match:
            return match.group("id")

        if self._cache.node_id is None and load_if_missing:
            self.fetch_metadata()
        return self._cache.node_id

    def name(self, *, load_if_missing: bool = True) -> Optional[str]:
        if self._cache.name is None and load_if_missing:
            self.fetch_metadata()
        return self._cache.name

    def type(self, *, load_if_missing: bool = True) -> Optional[LDataNodeType]:
        if self._cache.type is None and load_if_missing:
            self.fetch_metadata()
        return self._cache.type

    def size(self, *, load_if_missing: bool = True) -> Optional[int]:
        if self._cache.size is None and load_if_missing:
            self.fetch_metadata()
        return self._cache.size

    def content_type(self, *, load_if_missing: bool = True) -> Optional[str]:
        if self._cache.content_type is None and load_if_missing:
            self.fetch_metadata()
        return self._cache.content_type

    def is_dir(self, *, load_if_missing: bool = True) -> bool:
        return self.type(load_if_missing=load_if_missing) in _dir_types

    def iterdir(self) -> Iterator[Self]:
        """Yield LPaths objects contained within the directory.

        Should only be called on directories. Does not recursively list directories.

        Always makes a network request.
        """
        data = query_with_retry(
            gql.gql("""
            query LDataChildren($argPath: String!) {
                ldataResolvePathData(argPath: $argPath) {
                    finalLinkTarget {
                        type
                        childLdataTreeEdges(filter: { child: { removed: { equalTo: false } } }) {
                            nodes {
                                child {
                                    name
                                }
                            }
                        }
                    }
                }
            }"""),
            {"argPath": self.path},
        )["ldataResolvePathData"]

        if data is None:
            raise LatchPathError(f"no such Latch file or directory", {self.path})
        if data["finalLinkTarget"]["type"].lower() not in _dir_types:
            raise ValueError(f"not a directory: {self.path}")

        for node in data["finalLinkTarget"]["childLdataTreeEdges"]["nodes"]:
            yield LPath(urljoins(self.path, node["child"]["name"]))

    def mkdirp(self) -> None:
        node = _get_node_data(self.path, allow_resolve_to_parent=True).data[self.path]
        if node.exists():
            if node.type not in _dir_types:
                raise ValueError(f"{self.path} exists and is not a directory")
            return

        path = f"latch://{node.id}.node/{node.remaining}"
        path += "/" if not path.endswith("/") else ""
        query_with_retry(
            gql.gql("""
            mutation LDataMkdirP($path: String!) {
                ldataMkdirp(input: { argPath: $path }) {
                    bigInt
                }
            }
            """),
            {"path": path},
        )
        self._clear_cache()

    def rmr(self) -> None:
        """Recursively delete files at this instance's path.

        Throws LatchPathError if the path does not exist.

        Always makes a network request.
        """
        query_with_retry(
            gql.gql("""
            mutation LDataRmr($nodeId: BigInt!) {
                ldataRmr(input: { argNodeId: $nodeId }) {
                    clientMutationId
                }
            }
            """),
            {"nodeId": self.node_id()},
        )
        self._clear_cache()

    def copy_to(self, dst: "LPath") -> None:
        """Copy the file at this instance's path to the given destination.

        Args:
        dst: The destination LPath.
        show_summary: Whether to print a summary of the copy operation.
        """
        _remote_copy(self.path, dst.path, create_parents=True)

    def upload_from(self, src: Path, *, show_progress_bar: bool = False) -> None:
        """Upload the file at the given source to this instance's path.

        Args:
        src: The source path.
        show_progress_bar: Whether to show a progress bar during the upload.
        """
        _upload(
            os.fspath(src),
            self.path,
            progress=_Progress.tasks if show_progress_bar else _Progress.none,
            verbose=False,
            create_parents=True,
        )
        self._clear_cache()

    def download(
        self, dst: Optional[Path] = None, *, show_progress_bar: bool = False
    ) -> Path:
        """Download the file at this instance's path to the given destination.

        Args:
        dst: The destination path. If None, a temporary directory is created and the file is
            downloaded there. The temprary directory is deleted when the program exits.
        show_progress_bar: Whether to show a progress bar during the download.
        """
        if dst is None:
            global _download_idx
            tmp_dir = Path.home() / ".latch" / "lpath" / str(_download_idx)
            _download_idx += 1
            tmp_dir.mkdir(parents=True, exist_ok=True)
            atexit.register(lambda p: shutil.rmtree(p), tmp_dir)
            dst = tmp_dir / self.name()

        _download(
            self.path,
            dst,
            progress=_Progress.tasks if show_progress_bar else _Progress.none,
            verbose=False,
            confirm_overwrite=False,
        )
        return dst

    def __truediv__(self, other: object) -> "LPath":
        if not isinstance(other, str):
            return NotImplemented
        return LPath(urljoins(self.path, other))


class LPathTransformer(TypeTransformer[LPath]):
    _TYPE_INFO = BlobType(
        # todo(rahul): there is no way to know if the LPath is a file or directory
        # ahead to time, so just set dimensionality to SINGLE
        format="",
        dimensionality=BlobType.BlobDimensionality.SINGLE,
    )

    def __init__(self):
        super().__init__(name="lpath-transformer", t=LPath)

    def get_literal_type(self, t: Type[LPath]) -> LiteralType:
        return LiteralType(blob=self._TYPE_INFO)

    def to_literal(
        self,
        ctx: FlyteContext,
        python_val: LPath,
        python_type: Type[LPath],
        expected: LiteralType,
    ) -> Literal:
        return Literal(
            scalar=Scalar(
                blob=Blob(
                    uri=python_val.path, metadata=BlobMetadata(type=self._TYPE_INFO)
                )
            )
        )

    def to_python_value(
        self, ctx: FlyteContext, lv: Literal, expected_python_type: Type[LPath]
    ):
        return LPath(path=lv.scalar.blob.uri)


TypeEngine.register(LPathTransformer())



from pathlib import Path

from latch.types import LatchDir
from latch.types.metadata import LatchAuthor, SnakemakeFileParameter, SnakemakeMetadata

SnakemakeMetadata(
    output_dir=LatchDir("latch:///sample_output"),
    display_name="snakemake_tutorial_workflow",
    author=LatchAuthor(
        name="Kenneth",
    ),
    parameters={
        "samples": SnakemakeFileParameter(
            display_name="Sample Input Directory",
            description="A directory full of FastQ files",
            type=LatchDir,
            path=Path("data/samples"),
        ),
        "ref_genome": SnakemakeFileParameter(
            display_name="Indexed Reference Genome",
            description=(
                "A directory with a reference Fasta file and the 6 index files produced"
                " from `bwa index`"
            ),
            type=LatchDir,
            path=Path("genome"),
        ),
    },
)



import inspect
from dataclasses import is_dataclass
from textwrap import dedent
from typing import Callable, Dict, Union, get_args, get_origin

import click
from flytekit import workflow as _workflow
from flytekit.core.workflow import PythonFunctionWorkflow

from latch.types.metadata import LatchAuthor, LatchMetadata, LatchParameter
from latch_cli.utils import best_effort_display_name


def _generate_metadata(f: Callable) -> LatchMetadata:
    signature = inspect.signature(f)
    metadata = LatchMetadata(f.__name__, LatchAuthor())
    metadata.parameters = {
        param: LatchParameter(display_name=best_effort_display_name(param))
        for param in signature.parameters
    }
    return metadata


def _inject_metadata(f: Callable, metadata: LatchMetadata) -> None:
    if f.__doc__ is None:
        f.__doc__ = f"{f.__name__}\n\nSample Description"
    short_desc, long_desc = f.__doc__.split("\n", 1)
    f.__doc__ = f"{short_desc}\n{dedent(long_desc)}\n\n" + str(metadata)


# this weird Union thing is to ensure backwards compatibility,
# so that when users call @workflow without any arguments or
# parentheses, the workflow still serializes as expected
def workflow(
    metadata: Union[LatchMetadata, Callable]
) -> Union[PythonFunctionWorkflow, Callable]:
    if isinstance(metadata, Callable):
        f = metadata
        if f.__doc__ is None or "__metadata__:" not in f.__doc__:
            metadata = _generate_metadata(f)
            _inject_metadata(f, metadata)
        return _workflow(f)

    def decorator(f: Callable):
        signature = inspect.signature(f)
        wf_params = signature.parameters

        updated_params: Dict[str, LatchParameter] = {}
        for wf_param in wf_params:
            updated_params[wf_param] = (
                LatchParameter(display_name=best_effort_display_name(wf_param))
                if wf_param not in metadata.parameters
                else metadata.parameters[wf_param]
            )
        metadata.parameters = updated_params

        in_meta_not_in_wf = []
        for meta_param in metadata.parameters:
            if meta_param not in wf_params:
                in_meta_not_in_wf.append(meta_param)

        if len(in_meta_not_in_wf) > 0:
            error_str = (
                "Inconsistency detected between parameters in your `LatchMetadata`"
                " object and parameters in your workflow signature.\n\n"
                "The following parameters appear in your `LatchMetadata` object"
                " but not in your workflow signature:\n\n"
            )
            for meta_param in in_meta_not_in_wf:
                error_str += f"    \x1b[1m{meta_param}\x1b[22m\n"
            error_str += (
                "\nPlease resolve these inconsistencies and ensure that your"
                " `LatchMetadata` object and workflow signature have the same"
                " parameters."
            )
            click.secho(error_str, fg="red")
            raise click.exceptions.Exit(1)

        for name, meta_param in metadata.parameters.items():
            if meta_param.samplesheet is not True:
                continue

            annotation = wf_params[name].annotation

            origin = get_origin(annotation)
            args = get_args(annotation)
            valid = (
                origin is not None
                and issubclass(origin, list)
                and is_dataclass(args[0])
            )
            if not valid:
                click.secho(
                    f"parameter marked as samplesheet is not valid: {name} "
                    f"in workflow {f.__name__} must be a list of dataclasses",
                    fg="red",
                )
                raise click.exceptions.Exit(1)

        _inject_metadata(f, metadata)
        return _workflow(f)

    return decorator



"""The Latch SDK

A commandline toolchain to define and register serverless workflows with the
Latch platform.
"""

from latch.functions.messages import message
from latch.functions.operators import (
    combine,
    group_tuple,
    inner_join,
    latch_filter,
    left_join,
    outer_join,
    right_join,
)
from latch.resources.conditional import create_conditional_section
from latch.resources.map_tasks import map_task
from latch.resources.reference_workflow import workflow_reference
from latch.resources.tasks import (
    custom_task,
    custom_memory_optimized_task,
    large_gpu_task,
    large_task,
    medium_task,
    small_gpu_task,
    small_task,
)
from latch.resources.workflow import workflow



import click
import gql
from latch_sdk_gql.execute import execute

from latch.ldata._transfer.node import get_node_data as _get_node_data
from latch.ldata.path import LatchPathError
from latch_cli.services.cp.glob import expand_pattern


def rmr(
    remote_path: str,
    skip_confirmation: bool = False,
    no_glob: bool = False,
    verbose: bool = False,
):
    """Deletes an entity on Latch

    Args:
        remote_path:   A valid path to a remote destination, of the form

                            [latch://] [/] dir_1/dir_2/.../dir_n/entity_name,

                       where entity_name is the name of the entity to be removed.

    This function will remove the entity at the remote path specified recursively
    (like rm -r on POSIX systems), and will error if the remote path specified is
    invalid or if the entity doesn't exist.

    Example: ::

        rm("sample.txt") # sample.txt exists

            Removes the existing file sample.txt from Latch.

        rm("latch:///dir1/dir2") # dir1/dir2/ exists and is nonempty

            Removes the directory dir1/dir2 along with all of its contents.

        rm("/dir1/dir3/dir2/doesnt_exist.txt") # doesnt_exist.txt doesn't exist

            Will throw an error, as this operation tries to remove a file
            that doesn't exist.
    """
    to_remove = [remote_path] if no_glob else expand_pattern(remote_path)
    if len(to_remove) == 0:
        click.echo(
            f"Could not find any files that match pattern {remote_path}",
        )
        return

    msg = (
        "Remove the following file(s)?\n" + "\n".join(to_remove)
        if verbose
        else f"Remove {len(to_remove)} file/dir(s)?"
    )
    if not skip_confirmation and not click.confirm(msg):
        return

    try:
        node_data = _get_node_data(*to_remove).data
    except LatchPathError as e:
        click.secho(str(e), fg="red")
        raise click.exceptions.Exit(1) from e

    for path in to_remove:
        execute(
            gql.gql("""
                mutation LatchCLIRmr($argNodeId: BigInt!) {
                    ldataRmr(input: {argNodeId: $argNodeId}) {
                        clientMutationId
                    }
                }
            """),
            {"argNodeId": node_data[path].id},
        )
        if verbose:
            click.secho(f"Successfully deleted {path}.", fg="green")



from dataclasses import dataclass, field
from pathlib import Path
from traceback import walk_tb
from types import TracebackType
from typing import List, Optional, Type

from latch_cli.exceptions.cache import _code_cache
from latch_cli.exceptions.errors import _FlytekitError, _HandledError, _SyntaxError


@dataclass
class _Frame:
    filename: str
    lineno: int
    name: str
    locals_: dict

    def read(self, line_interval: int = 3) -> str:
        cache_hit = _code_cache.get((self.filename, self.lineno))
        if cache_hit is not None:
            return cache_hit

        span: List[str] = []
        with open(self.filename, "r") as code_file:
            for i, line in enumerate(code_file.readlines()):
                if i in range(
                    self.lineno - line_interval,
                    self.lineno + line_interval + 1,
                ):
                    span.append((i + 1, line))

        _code_cache[(self.filename, self.lineno)] = span

        return span


@dataclass
class _Stack:
    exc_type: str
    exc_value: str
    is_cause: bool = False
    error: Optional[_HandledError] = None
    frames: List[_Frame] = field(default_factory=list)

    def pretty_print(self, max_frames: int = 20):
        render_full_idx = len(self.frames) - max_frames
        for i, frame in enumerate(self.frames):
            if i < render_full_idx:
                print(f"{frame.filename}:{frame.lineno} in {frame.name}")
                if i > 0 and i == render_full_idx - 1:
                    print(f"{i} frames collapsed...")
            else:
                print()
                print("--Traceback (most recent call last)--")
                print(f"{frame.filename}:{frame.lineno} in {frame.name}")

                for lineno, line in frame.read():
                    if lineno == frame.lineno:
                        print(f"> {lineno}| {line}", end="")
                    else:
                        print(f"  {lineno}| {line}", end="")


class _Traceback:
    def __init__(self, exc_type, exc_value, traceback):
        self.stacks = self.extract_stack(exc_type, exc_value, traceback)

    def extract_stack(
        self,
        exc_type: Type[BaseException],
        exc_value: BaseException,
        traceback: Optional[TracebackType],
    ) -> List[_Stack]:
        stacks: List[_Stack] = []
        while True:
            stack = _Stack(
                exc_type=exc_type.__name__,
                exc_value=exc_value,
            )
            stacks.append(stack)

            if isinstance(exc_value, SyntaxError):
                stack.error = _SyntaxError(
                    filename=exc_value.filename or "?",
                    lineno=exc_value.lineno or 0,
                    offset=exc_value.offset or 0,
                    text=exc_value.text or "",
                    end_lineno=exc_value.end_lineno or 0,
                    end_offset=exc_value.end_offset or 0,
                )
            elif isinstance(exc_value, _FlytekitError):
                stack.error = _FlytekitError()

            for frame_summary, line_no in walk_tb(traceback):
                filename = frame_summary.f_code.co_filename

                # Frames with <frozen importlib._bootstrap> have no real debugging
                # information and just add noise to the traceback.
                if filename.startswith("<"):
                    continue
                if filename:
                    filename = str(Path(filename).resolve())

                stack.frames.append(
                    _Frame(
                        filename=filename or "?",
                        lineno=line_no,
                        name=frame_summary.f_code.co_name,
                        locals_=frame_summary.f_locals.copy(),
                    )
                )

            # Explicitly chained - raise ... from
            # https://peps.python.org/pep-3134/#motivation
            cause = getattr(exc_value, "__cause__", None)
            if cause is not None:
                stack.is_cause = True
                exc_type = cause.__class__
                exc_value = cause
                traceback = cause.__traceback__
                continue

            # Implicity chained
            context = exc_value.__context__
            if context and not getattr(exc_value, "__suppress_context__", False):
                stack.is_cause = False
                exc_type = context.__class__
                exc_value = context
                traceback = context.__traceback__
                continue

            break

        return stacks[::-1]

    def pretty_print(self):
        for i, stack in enumerate(self.stacks):
            stack.pretty_print()
            print(f"\n{stack.exc_type}: {stack.exc_value}")

            if i < len(self.stacks) - 1:
                if stack.is_cause:
                    print(
                        "\nThe above exception was the direct cause of the following"
                        " exception:\n"
                    )
                else:
                    print(
                        "\nDuring handling of the above exception, another exception"
                        " occurred:\n"
                    )



import os
import stat
import sys
from datetime import datetime
from pathlib import Path
from typing import Dict, List, Optional, Tuple

import click
import dateutil.parser as dp
import gql
from gql.transport.exceptions import TransportQueryError
from latch_sdk_gql.execute import JsonValue, execute

import latch.ldata._transfer.upload as _upl
from latch_cli.utils.path import is_remote_path


def upload_file(src: Path, dest: str):
    start = _upl.start_upload(src, dest)
    if start is None:
        return

    parts: List[_upl.CompletedPart] = []
    for idx, url in enumerate(start.urls):
        parts.append(
            _upl.upload_file_chunk(
                src,
                url,
                idx,
                start.part_size,
            )
        )

    _upl.end_upload(dest, start.upload_id, parts)


def check_src(p: Path, *, indent: str = "") -> Optional[Tuple[Path, os.stat_result]]:
    try:
        p_stat = os.stat(p)
    except FileNotFoundError:
        click.secho(indent + f"`{p}`: no such file or directory", fg="red", bold=True)
        return

    if not stat.S_ISREG(p_stat.st_mode) and not stat.S_ISDIR(p_stat.st_mode):
        click.secho(indent + f"`{p}`: not a regular file", fg="red", bold=True)
        return

    return (p, p_stat)


def sync_rec(
    srcs: Dict[str, Tuple[Path, os.stat_result]],
    dest: str,
    *,
    delete: bool,
    level: int = 0,
):
    # rsync never deletes from the top level destination
    delete_effective = delete and level > 0
    indent = "  " * level

    try:
        query = """
            query LatchCLISync($argPath: String! ${name_filter_arg}) {
                ldataResolvePathData(argPath: $argPath) {
                    finalLinkTarget {
                        type
                        childLdataTreeEdges(
                            filter: {
                                child: {
                                    removed: {equalTo: false},
                                    pending: {equalTo: false},
                                    copiedFrom: {isNull: true}
                                    ${name_filter}
                                }
                            }
                        ) {
                            nodes {
                                child {
                                    id
                                    name
                                    finalLinkTarget {
                                        type
                                        ldataNodeEvents(
                                            condition: {type: INGRESS},
                                            orderBy: TIME_DESC,
                                            first: 1
                                        ) {
                                            nodes {
                                                time
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        """

        args: JsonValue = {"argPath": dest, "nameFilter": []}
        if not delete_effective:
            query = query.replace("${name_filter_arg}", ", $nameFilter: [String!]")
            query = query.replace("${name_filter}", ", name: {in: $nameFilter}")
            args["nameFilter"] = list(srcs.keys())
        else:
            query = query.replace("${name_filter_arg}", "")
            query = query.replace("${name_filter}", "")

        resolve_data = execute(
            gql.gql(query),
            args,
        )["ldataResolvePathData"]

        dest_data = None
        if resolve_data is not None:
            dest_data = resolve_data["finalLinkTarget"]
    except TransportQueryError as e:
        if e.errors is None or len(e.errors) == 0:
            raise

        msg: str = e.errors[0]["message"]

        raise

    if len(srcs) == 0:
        if dest_data is not None:
            if dest_data["type"] != "DIR":
                click.secho(
                    indent + f"`{dest}` is in the way of a directory",
                    fg="red",
                )
                return

            click.secho(indent + "Empty directory", dim=True)
            return

        if not dest[-1] == "/":
            dest += "/"

        click.secho(indent + "Creating empty directory", fg="bright_blue")
        execute(
            gql.gql("""
                mutation LatchCLISyncMkdir($argPath: String!) {
                    ldataMkdirp(input: {argPath: $argPath}) {
                        clientMutationId
                    }
                }
            """),
            {"argPath": dest},
        )
        return

    if (
        (len(srcs) > 1 or stat.S_ISDIR(list(srcs.values())[0][1].st_mode))
        and dest_data is not None
        and dest_data["type"] not in {"DIR", "ACCOUNT_ROOT"}
    ):
        click.secho(f"`{dest}` is not a directory", fg="red", bold=True)
        click.secho("\nOnly a single file can be synced with a file", fg="red")
        sys.exit(1)

    if dest_data is not None and dest_data["type"] not in {"DIR", "ACCOUNT_ROOT"}:
        # todo(maximsmol): implement
        click.secho(
            "Syncing single files is currently not supported", bold=True, fg="red"
        )
        sys.exit(1)

    dest_children_by_name = (
        {
            x["name"]: x
            for x in (raw["child"] for raw in dest_data["childLdataTreeEdges"]["nodes"])
        }
        if dest_data is not None
        else {}
    )

    for name, (p, p_stat) in srcs.items():
        is_dir = stat.S_ISDIR(p_stat.st_mode)

        child = dest_children_by_name.get(name)
        child_dest = f"{dest}/{name}"

        skip = False
        verb = "Uploading"
        reason = "new"
        if child is not None:
            flt = child["finalLinkTarget"]
            if flt["type"] == "DIR" and not is_dir:
                # todo(maximsmol): confirm? pre-check?
                click.secho(
                    indent + f"`{dest}` is in the way of a file",
                    fg="red",
                )
                continue

            if flt["type"] != "DIR" and is_dir:
                # todo(maximsmol): confirm? pre-check?
                click.secho(
                    indent + f"`{dest}` is in the way of a directory",
                    fg="red",
                )
                continue

            if flt["type"] == "OBJ":
                remote_mtime = dp.isoparse(flt["ldataNodeEvents"]["nodes"][0]["time"])

                local_mtime = datetime.fromtimestamp(p_stat.st_mtime).astimezone()
                if remote_mtime == local_mtime:
                    verb = "Skipping"
                    reason = "unmodified"
                    skip = True
                elif remote_mtime > local_mtime:
                    verb = "Skipping"
                    reason = "older"
                    skip = True
                else:
                    verb = "Uploading"
                    reason = "updated"
            else:
                reason = "existing"

        if verb == "Uploading" and is_dir:
            verb = "Syncing"

        fg = "bright_blue"
        dim = None
        if verb == "Skipping":
            fg = None
            dim = True

        click.echo(
            click.style(
                indent + verb + " ",
                fg=fg,
                dim=dim,
            )
            + click.style(
                reason,
                underline=True,
                fg=fg,
                dim=dim,
            )
            + click.style(
                ": ",
                fg=fg,
                dim=dim,
            )
            + click.style(
                str(p)
                + ("" if not is_dir else "/")
                + ("" if skip else click.style(" -> ", dim=True) + child_dest),
                dim=dim,
            )
        )
        if skip:
            continue

        if is_dir:
            sub_srcs: Dict[str, Tuple[Path, os.stat_result]] = {}
            for x in p.iterdir():
                res = check_src(x, indent=indent + "  ")
                if res is None:
                    # todo(maximsmol): pre-check or confirm?
                    continue

                sub_srcs[x.name] = res
            sync_rec(sub_srcs, child_dest, delete=delete, level=level + 1)
            continue

        # todo(maximsmol): upload in parallel?
        upload_file(p, child_dest)

    if delete_effective:
        for name, child in dest_children_by_name.items():
            child_dest = f"{dest}/{name}"
            if name in srcs:
                continue

            click.echo(
                indent + click.style("Removing extraneous: ", fg="yellow") + child_dest
            )
            execute(
                gql.gql("""
                    mutation LatchCLISyncRemove($argNodeId: BigInt!) {
                        ldataRmr(input: {argNodeId: $argNodeId}) {
                            clientMutationId
                        }
                    }
                """),
                {"argNodeId": child["id"]},
            )


def sync(
    srcs_raw: List[str],
    dest: str,
    *,
    delete: bool,
    ignore_unsyncable: bool,
):
    if not is_remote_path(dest):
        click.secho(
            f"`{dest}`: only local -> remote sync is supported", fg="red", bold=True
        )
        raise click.exceptions.Exit(1)

    srcs: Dict[str, Tuple[Path, os.stat_result]] = {}
    have_errors = False
    for x in srcs_raw:
        if is_remote_path(x):
            click.secho(
                f"`{x}`: only local -> remote sync is supported", fg="red", bold=True
            )
            have_errors = True
            continue

        p = Path(x)
        res = check_src(p)
        if res is None:
            have_errors = True
            continue

        srcs[p.name] = res

    if len(srcs) == 0:
        click.secho(
            "\nAll source paths were skipped due to errors", fg="red", bold=True
        )
        raise click.exceptions.Exit(1)

    if have_errors:
        # todo(maximsmol): do we want to precheck recursively?
        click.secho("\nSome source paths will be skipped due to errors", fg="red")

        if not ignore_unsyncable:
            if not click.confirm(click.style(f"Proceed?", fg="red")):
                sys.exit(1)
        else:
            click.secho(
                "Proceeding due to " + click.style("`--ignore-unsyncable`", bold=True),
                fg="yellow",
            )
        click.echo()

    sync_rec(srcs, dest, delete=delete)



"""Service to login."""


from typing import Optional

import click
from latch_sdk_config.latch import config
from latch_sdk_config.user import user_config


def login(connection: Optional[str] = None) -> str:
    """Authenticates a user with Latch and persists an access token.

    Kicks off the three-legged OAuth2.0 flow outlined in `this RFC`_.  The logic
    scaffolding this flow and detailed documentation about it can be found in
    the `latch.auth` package.

    The user will be redirected to a browser and prompted to login. This
    function meanwhile spins up a callback server on a separate thread that will
    be hit when the browser login is successful with an access token.

    .. _this RFC:
        https://datatracker.ietf.org/doc/html/rfc6749
    """
    if _browser_available() is False:
        token: str = click.prompt(
            f"Go to `{config.console_routes.developer}` and copy your API Key here",
            type=str,
        )
        token = token.strip()
        user_config.update_token(token)

        return token

    from latch_cli.auth import PKCE, CSRFState, OAuth2
    from latch_cli.constants import oauth2_constants

    with PKCE() as pkce:
        with CSRFState() as csrf_state:
            oauth2_flow = OAuth2(pkce, csrf_state, oauth2_constants)
            auth_code = oauth2_flow.authorization_request(connection)
            jwt = oauth2_flow.access_token_request(auth_code)

            # Exchange JWT from Auth0 for a persistent token issued by
            # LatchBio.
            access_jwt = _auth0_jwt_for_access_jwt(jwt)
            user_config.update_token(access_jwt)

            return access_jwt


def _browser_available() -> bool:
    """Returns true if browser available for login flow.

    Takes advantage of browser searching logic for many platforms written
    `here`_.

    .. _here:
        https://github.com/python/cpython/blob/3a2b89580ded72262fbea0f7ad24096a90c42b9c/Lib/webbrowser.py#L38
    """
    import webbrowser

    try:
        browser = webbrowser.get()
        if browser is not None:
            return True
    except Exception:
        pass
    return False


def _auth0_jwt_for_access_jwt(token) -> str:
    """Requests a LatchBio minted (long-lived) acccess JWT.

    Uses an Auth0 token to authenticate the user.
    """
    import latch_cli.tinyrequests as tinyrequests

    headers = {
        "Authorization": f"Bearer {token}",
    }

    url = config.api.user.jwt

    response = tinyrequests.post(url, headers=headers)

    resp = response.json()
    try:
        jwt = resp["jwt"]
    except KeyError as e:
        raise ValueError(
            f"Malformed response from request for access token {resp}"
        ) from e

    return jwt



import secrets

from latch_cli.auth.utils import _base64url_encode


class CSRFState:
    """Context manager to hold state preventing CSRF attacks.

    Outlined in detail `here`_, this object holds a state string as a
    client-side protection against CSRF attacks.

    Used in conjunction with PKCE verifier/challenge protection, which is a
    server-side security mechanism.

    Example usage: ::

        with CSRFState() as csrf:
            oauth2_flow = OAuth2(..., csrf, ...)

    .. _here:
        https://spring.io/blog/2011/11/30/cross-site-request-forgery-and-oauth2
    """

    def __init__(self):
        self.state = _base64url_encode(secrets.token_bytes(32))

    def __enter__(self, *args):
        return self

    def __exit__(self, *args): ...



import builtins
import contextlib
import os
import random
import string
import sys
import tempfile
from dataclasses import dataclass
from pathlib import Path
from types import ModuleType
from typing import Iterator, List, Optional

import docker
import paramiko
from flytekit.core.context_manager import FlyteContext, FlyteContextManager
from flytekit.core.data_persistence import FileAccessProvider
from flytekit.tools import module_loader

from latch_cli.constants import latch_constants


@dataclass
class RemoteConnInfo:
    ip: str
    username: str
    jump_key_path: Path
    ssh_key_path: Path


@contextlib.contextmanager
def _add_sys_paths(paths: List[Path]) -> Iterator[None]:
    paths = [os.fspath(p) for p in paths]
    try:
        for p in paths:
            sys.path.insert(0, p)
        yield
    finally:
        for p in paths:
            sys.path.remove(p)


def _import_flyte_objects(paths: List[Path], module_name: str = "wf"):
    with _add_sys_paths(paths):

        class FakeModule(ModuleType):
            def __getattr__(self, key):
                # This value is designed to be used in the execution of top
                # level code without throwing errors but need not do anything.
                # Here are some cases where this value can be used:
                #
                # ```
                # from foo import bar
                #
                # # Referencing an attribute from an attribute
                # a = bar.x
                #
                # # Calling an attribute
                # b = bar()
                #
                # # Subclassing from an attribute
                # class C(bar):
                # ```
                #     ...
                class FakeAttr:
                    def __new__(*args, **kwargs):
                        return None

                return FakeAttr

            __all__ = []

        real_import = builtins.__import__

        def fake_import(name, globals=None, locals=None, fromlist=(), level=0):
            try:
                return real_import(
                    name,
                    globals=globals,
                    locals=locals,
                    fromlist=fromlist,
                    level=level,
                )
            except (ModuleNotFoundError, AttributeError) as e:
                return FakeModule(name)

        # Temporary ctx tells lytekit to skip local execution when
        # inspecting objects
        fap = FileAccessProvider(
            local_sandbox_dir=tempfile.mkdtemp(prefix="foo"),
            raw_output_prefix="bar",
        )
        tmp_context = FlyteContext(fap, inspect_objects_only=True)

        FlyteContextManager.push_context(tmp_context)
        builtins.__import__ = fake_import

        mods = list(module_loader.iterate_modules([module_name]))

        builtins.__import__ = real_import
        FlyteContextManager.pop_context()

        return mods


def _construct_dkr_client(ssh_host: Optional[str] = None):
    """Try many methods of establishing valid connection with client.

    This was helpful -
    https://github.com/docker/docker-py/blob/a48a5a9647761406d66e8271f19fab7fa0c5f582/docker/utils/utils.py#L321

    If `ssh_host` is passed, we attempt to make a connection with a remote
    machine.
    """

    def _from_env():
        host = environment.get("DOCKER_HOST")

        # empty string for cert path is the same as unset.
        cert_path = environment.get("DOCKER_CERT_PATH")
        if cert_path == "":
            cert_path = None

        # empty string for tls verify counts as "false".
        # Any value or 'unset' counts as true.
        tls_verify = environment.get("DOCKER_TLS_VERIFY") != ""

        enable_tls = tls_verify or cert_path is not None

        dkr_client = None
        try:
            if not enable_tls:
                dkr_client = docker.APIClient(host)
            else:
                if not cert_path:
                    cert_path = os.path.join(os.path.expanduser("~"), ".docker")

                tls_config = docker.tls.TLSConfig(
                    client_cert=(
                        os.path.join(cert_path, "cert.pem"),
                        os.path.join(cert_path, "key.pem"),
                    ),
                    ca_cert=os.path.join(cert_path, "ca.pem"),
                    verify=tls_verify,
                )
                dkr_client = docker.APIClient(host, tls=tls_config)

        except docker.errors.DockerException as de:
            raise OSError(
                "Unable to establish a connection to Docker. Make sure that"
                " Docker is running and properly configured before attempting"
                " to register a workflow."
            ) from de

        return dkr_client

    if ssh_host is not None:
        try:
            return docker.APIClient(ssh_host, version="1.41")
        except docker.errors.DockerException as de:
            raise OSError(
                f"Unable to establish a connection to remote docker host {ssh_host}."
            ) from de

    environment = os.environ
    host = environment.get("DOCKER_HOST")
    if host is not None and host != "":
        return _from_env()
    else:
        try:
            # TODO: platform specific socket defaults
            return docker.APIClient(base_url="unix://var/run/docker.sock")
        except docker.errors.DockerException as de:
            raise OSError(
                "Docker is not running. Make sure that"
                " Docker is running before attempting to register a workflow."
            ) from de


def _construct_ssh_client(
    remote_conn_info: RemoteConnInfo, *, use_gateway: bool = True
) -> paramiko.SSHClient:
    if use_gateway:
        gateway = paramiko.SSHClient()
        gateway.load_system_host_keys()
        gateway.set_missing_host_key_policy(paramiko.MissingHostKeyPolicy)

        gateway_pkey = paramiko.PKey.from_path(
            path=str(remote_conn_info.jump_key_path.resolve())
        )

        gateway.connect(
            latch_constants.jump_host,
            username=latch_constants.jump_user,
            pkey=gateway_pkey,
        )

        gateway_transport = gateway.get_transport()
        if gateway_transport is None:
            raise ConnectionError("unable to create connection to jump host")

        sock = gateway_transport.open_channel(
            kind="direct-tcpip",
            dest_addr=(remote_conn_info.ip, 22),
            src_addr=("", 0),
        )
    else:
        sock = None

    pkey = paramiko.PKey.from_path(path=str(remote_conn_info.ssh_key_path.resolve()))

    ssh = paramiko.SSHClient()
    ssh.load_system_host_keys()
    ssh.set_missing_host_key_policy(paramiko.MissingHostKeyPolicy)
    ssh.connect(
        remote_conn_info.ip,
        username=remote_conn_info.username,
        sock=sock,
        pkey=pkey,
    )

    transport = ssh.get_transport()
    if transport is None:
        raise ConnectionError(
            "unable to create connection from jump host to centromere deployment"
        )

    # (kenny) Equivalent of OpenSSH configuration `ServerAliveInterval`
    # No analogue for `ServerAliveCountMax` in paramiko I could find.
    transport.set_keepalive(latch_constants.centromere_keepalive_interval)

    return ssh


class MaybeRemoteDir:
    """A temporary directory that exists locally or on a remote machine."""

    def __init__(self, ssh_client: Optional[paramiko.SSHClient] = None):
        self.ssh_client = ssh_client

    def __enter__(self):
        return self.create()

    def __exit__(self, exc_type, exc_value, tb):
        self.cleanup()

    def create(self):
        if self.ssh_client is None:
            self._tempdir = tempfile.TemporaryDirectory()
            return str(Path(self._tempdir.name).resolve())

        td = build_random_string()

        self._tempdir = f"/tmp/{td}"

        self.ssh_client.exec_command(f"mkdir {self._tempdir}")
        return self._tempdir

    def cleanup(self):
        if self.ssh_client is not None:
            self.ssh_client.exec_command(f"rm -rf {self._tempdir}")
            return

        if self._tempdir is not None and not isinstance(self._tempdir, str):
            self._tempdir.cleanup()


def build_random_string(k: int = 8) -> str:
    return "".join(
        random.choices(
            string.ascii_uppercase + string.ascii_lowercase + string.digits, k=k
        )
    )



import subprocess
from pathlib import Path

from latch.functions.messages import message
from latch.resources.tasks import medium_task
from latch.types.directory import LatchDir, LatchOutputDir
from latch.types.file import LatchFile


@medium_task
def run_nfcore_fetchngs(ids: LatchFile, output_directory: LatchOutputDir) -> LatchDir:
    outdir = Path("/root/outputs").resolve()
    outdir.mkdir(exist_ok=True)

    nfcore_cmd = [
        "nextflow",
        "run",
        "nf-core/fetchngs",
        "--input",
        ids.local_path,
        "--outdir",
        str(outdir),
        "-profile",
        "docker",
    ]

    try:
        subprocess.run(nfcore_cmd, check=True, capture_output=True)
    except subprocess.CalledProcessError as e:
        stderr = e.stderr.decode("utf-8")
        if stderr:
            message(
                "error",
                {"title": "Blastp Failed", "body": f"Stderr: {stderr}"},
            )
        print(stderr)
        raise e

    return LatchDir(path=str(outdir), remote_path=output_directory.remote_path)



from enum import Enum
from typing import IO, Any, Optional, Sequence, Tuple, Type

from click import Choice, Command, Context, Group, HelpFormatter, Parameter, echo, style
from click._compat import get_text_stderr


class EnumChoice(Choice):
    def __init__(self, choices: Type[Enum], case_sensitive: bool = True):
        self.enum = choices
        return super().__init__(choices._member_names_, case_sensitive)

    def convert(
        self, value: Any, param: Optional[Parameter], ctx: Optional[Context]
    ) -> Any:
        ret = super().convert(value, param, ctx)
        return self.enum(ret)

    # todo(ayush): override `shell_complete` once we support it


class ColoredHelpFormatter(HelpFormatter):
    def write_usage(
        self, prog: str, args: str = "", prefix: Optional[str] = None
    ) -> None:
        if prefix is None:
            prefix = "Usage: "

        prefix = style(prefix, fg="bright_blue")
        return super().write_usage(prog, args, prefix)

    def write_heading(self, heading: str):
        return super().write_heading(style(heading, fg="bright_blue"))

    def write_dl(
        self, rows: Sequence[Tuple[str, str]], col_max: int = 30, col_spacing: int = 2
    ) -> None:
        return super().write_dl(
            [(style(x[0], fg="bright_green", italic=True), x[1]) for x in rows],
            col_max,
            col_spacing,
        )


class LatchCommand(Command):
    def format_epilog(self, ctx: Context, formatter: HelpFormatter) -> None:
        formatter.write_paragraph()
        formatter.write_text(
            style("See " + style("https://docs.latch.bio/", underline=True), dim=True)
            + style(" for manuals, tutorials, examples, and an API reference", dim=True)
        )


class LatchGroup(LatchCommand, Group): ...


def colored_exception_show(self, file: Optional[IO] = None) -> None:
    if file is None:
        file = get_text_stderr()

    echo(
        style(f"Error: {style(self.format_message(), bold=True)}", fg="red"), file=file
    )


def colored_usage_error_show(self, file: Optional[IO] = None) -> None:
    if file is None:
        file = get_text_stderr()
    color = None
    hint = ""
    if self.ctx is not None and self.ctx.command.get_help_option(self.ctx) is not None:
        hint = (
            "Try "
            + style(
                f"'{self.ctx.command_path} {self.ctx.help_option_names[0]}'",
                bold=True,
            )
            + " for help."
        )
        hint = f"{hint}\n"
    if self.ctx is not None:
        color = self.ctx.color
        echo(f"{self.ctx.get_usage()}\n{hint}", file=file, color=color)
    echo(
        style(f"Error: {style(self.format_message(), bold=True)}", fg="red"),
        file=file,
        color=color,
    )


def patch():
    import click

    click.Context.formatter_class = ColoredHelpFormatter

    old_group = click.group
    click.group = lambda *args, **kwargs: old_group(*args, **kwargs, cls=LatchGroup)
    Group.command_class = LatchCommand

    click.ClickException.show = colored_exception_show
    click.UsageError.show = colored_usage_error_show


class AnsiCodes:
    full_reset = "\x1b[0m"

    color = "\x1b[38;5;39m"
    reset_color = "\x1b[39m"

    bold = "\x1b[1m"
    reset_bold = "\x1b[22m"

    underline = "\x1b[4m"
    no_underline = "\x1b[24m"

    # todo(maximsmol): use in supported terminals?
    url_href = "\x1b]8;;"
    url_name = "\x1b\\"
    url_end = "\x1b]8;;\x1b\\"


def bold(s: str) -> str:
    return f"{AnsiCodes.bold}{s}{AnsiCodes.reset_bold}"


def underline(s: str) -> str:
    return f"{AnsiCodes.underline}{s}{AnsiCodes.no_underline}"


def color(s: str, *, color: str = AnsiCodes.color):
    return f"{color}{s}{AnsiCodes.reset_color}"



try:
    from typing import get_args, get_origin
except ImportError:
    from typing_extensions import get_args, get_origin

import enum
import json
import keyword
import typing
from typing import Optional

import google.protobuf.json_format as gpjson
from flyteidl.core.literals_pb2 import Literal as _Literal
from flyteidl.core.types_pb2 import LiteralType as _LiteralType
from flytekit.models.literals import Literal
from flytekit.models.types import LiteralType

from latch.types.directory import LatchDir
from latch.types.file import LatchFile
from latch_cli.services.launch import _get_workflow_interface
from latch_cli.utils import retrieve_or_login


class _Unsupported: ...


_simple_table = {
    0: type(None),
    1: int,
    2: float,
    3: str,
    4: bool,
    5: _Unsupported,
    6: _Unsupported,
    7: _Unsupported,
    8: _Unsupported,
    9: _Unsupported,
}

_primitive_table = {
    type(None): None,
    int: 0,
    float: 0.0,
    str: "foo",
    bool: False,
}

# TODO(ayush): fix this to
# (1) support records,
# (2) support fully qualified workflow names,
# (note from kenny) - pretty sure you intend to support the opposite,
# fqn are supported by default, address when you get to this todo
# (3) show a message indicating the generated filename,
# (4) optionally specify the output filename


def get_params(wf_name: str, wf_version: Optional[str] = None):
    """Constructs a parameter map for a workflow given its name and an optional
    version.

    This function creates a python parameter file that can be used by `launch`.
    You can specify the specific parameters by editing the file, and then launch
    an execution on Latch using those parameters with `launch`.

    Args:
        wf_name: The unique name of the workflow.
        wf_version: An optional workflow version. If this argument is not given,
            `get_params` will default to generating a parameter map of the most
            recent version of the workflow.

    Example:
        >>> get_params("wf.__init__.alphafold_wf")
            # creates a file called `wf.__init__.alphafold_wf.params.py` that
            # contains a template parameter map.
    """

    token = retrieve_or_login()
    wf_id, wf_interface, wf_default_params = _get_workflow_interface(
        token, wf_name, wf_version
    )

    params = {}
    wf_vars = wf_interface["variables"]
    default_wf_vars = wf_default_params["parameters"]
    for key, value in wf_vars.items():
        try:
            description_json = json.loads(value["description"])
            param_name = description_json["name"]
        except (json.decoder.JSONDecodeError, KeyError) as e:
            raise ValueError(
                f"Parameter description json for workflow {wf_name} is malformed"
            ) from e

        literal_type_json = value["type"]
        literal_type = gpjson.ParseDict(literal_type_json, _LiteralType())

        python_type = _guess_python_type(
            LiteralType.from_flyte_idl(literal_type), param_name
        )

        default = True
        if default_wf_vars[param_name].get("required") is not True:
            literal_json = default_wf_vars[param_name].get("default")
            literal = gpjson.ParseDict(literal_json, _Literal())
            val = _guess_python_val(Literal.from_flyte_idl(literal), python_type)
        else:
            default = False
            val = _best_effort_default_val(python_type)

        params[param_name] = (python_type, val, default)

    import_statements = {
        LatchFile: "from latch.types import LatchFile",
        LatchDir: "from latch.types import LatchDir",
        enum.Enum: "from enum import Enum",
    }

    import_types = []
    enum_literals = []
    param_map_str = ""
    param_map_str += "\nparams = {"
    param_map_str += f'\n    "_name": "{wf_name}", # Don\'t edit this value.'
    for param_name, value in params.items():
        python_type, python_val, default = value

        # Check for imports.

        def _check_and_import(python_type: typing.T):
            if python_type in import_statements and python_type not in import_types:
                import_types.append(python_type)

        def _handle_enum(python_type: typing.T):
            if type(python_type) is enum.EnumMeta:
                if enum.Enum not in import_types:
                    import_types.append(enum.Enum)

                variants = python_type._variants
                name = python_type._name

                _enum_literal = f"class {name}(Enum):"
                for variant in variants:
                    if variant in keyword.kwlist:
                        variant_name = f"_{variant}"
                    else:
                        variant_name = variant
                    _enum_literal += f"\n    {variant_name} = '{variant}'"
                enum_literals.append(_enum_literal)

        # Parse collection, union types for potential imports and dependent
        # objects, eg. enum class construction.
        if get_origin(python_type) is not None:
            if get_origin(python_type) is list:
                _check_and_import(get_args(python_type)[0])
                _handle_enum(get_args(python_type)[0])
            elif get_origin(python_type) is typing.Union:
                for variant in get_args(python_type):
                    _check_and_import(variant)
                    _handle_enum(variant)
        else:
            _check_and_import(python_type)
            _handle_enum(python_type)

        python_val, python_type = _get_code_literal(python_val, python_type)

        if default is True:
            default = "DEFAULT. "
        else:
            default = ""

        param_map_str += f'\n    "{param_name}": {python_val}, # {default}{python_type}'
    param_map_str += "\n}"

    with open(f"{wf_name}.params.py", "w") as f:
        f.write(
            f'"""Run `latch launch {wf_name}.params.py` to launch this workflow"""\n'
        )

        for t in import_types:
            f.write(f"\n{import_statements[t]}")
        for e in enum_literals:
            f.write(f"\n\n{e}\n")

        f.write("\n")
        f.write(param_map_str)


def _get_code_literal(python_val: any, python_type: typing.T):
    """Construct value that is executable python when templated into a code
    block."""

    if python_type is str or (type(python_val) is str and str in get_args(python_type)):
        return f'"{python_val}"', python_type

    if type(python_type) is enum.EnumMeta:
        name = python_type._name
        return python_val, f"<enum '{name}'>"

    if get_origin(python_type) is typing.Union:
        variants = get_args(python_type)
        type_repr = "typing.Union["
        for i, variant in enumerate(variants):
            if i < len(variants) - 1:
                delimiter = ", "
            else:
                delimiter = ""
            type_repr += f"{_get_code_literal(python_val, variant)[1]}{delimiter}"
        type_repr += "]"
        return python_val, type_repr

    if get_origin(python_type) is list:
        if python_val is None:
            _, type_repr = _get_code_literal(None, get_args(python_type)[0])
            return None, f"typing.List[{type_repr}]"
        else:
            collection_literal = "["
            if len(python_val) > 0:
                for i, item in enumerate(python_val):
                    item_literal, type_repr = _get_code_literal(
                        item, get_args(python_type)[0]
                    )

                    if i < len(python_val) - 1:
                        delimiter = ","
                    else:
                        delimiter = ""

                    collection_literal += f"{item_literal}{delimiter}"
            else:
                list_t = get_args(python_type)[0]
                _, type_repr = _get_code_literal(
                    _best_effort_default_val(list_t), list_t
                )

            collection_literal += "]"
            return collection_literal, f"typing.List[{type_repr}]"

    return python_val, python_type


def _guess_python_val(literal: _Literal, python_type: typing.T):
    """Transform flyte literal value to native python value."""

    if literal.scalar is not None:
        if literal.scalar.none_type is not None:
            return None

        if literal.scalar.primitive is not None:
            primitive = literal.scalar.primitive

            if primitive.string_value is not None:
                if type(python_type) is enum.EnumMeta:
                    return f"{python_type._name}.{str(primitive.string_value)}"
                return str(primitive.string_value)

            if primitive.integer is not None:
                return int(primitive.integer)
            if primitive.float_value is not None:
                return float(primitive.float_value)
            if primitive.boolean is not None:
                return bool(primitive.boolean)

        if literal.scalar.blob is not None:
            blob = literal.scalar.blob
            dim = blob.metadata.type.dimensionality
            if dim == 0:
                return LatchFile(blob.uri)
            else:
                return LatchDir(blob.uri)

    # collection
    if literal.collection is not None:
        p_list = []
        for item in literal.collection.literals:
            p_list.append(_guess_python_val(item, get_args(python_type)[0]))
        return p_list

    # sum

    # enum

    raise NotImplementedError(
        f"The flyte literal {literal} cannot be transformed to a python type."
    )


def _guess_python_type(literal: LiteralType, param_name: str):
    """Transform flyte type literal to native python type."""

    if literal.simple is not None:
        return _simple_table[literal.simple]

    if literal.collection_type is not None:
        return typing.List[_guess_python_type(literal.collection_type, param_name)]

    if literal.blob is not None:
        # flyteidl BlobType message for reference:
        #   enum BlobDimensionality {
        #       SINGLE = 0;
        #       MULTIPART = 1;
        #   }

        dim = literal.blob.dimensionality
        if dim == 0:
            return LatchFile
        else:
            return LatchDir

    if literal.union_type is not None:
        variant_types = [
            _guess_python_type(variant, param_name)
            for variant in literal.union_type.variants
        ]

        # Trying to directly construct set of types will throw error if list is
        # included as 'list' is not hashable.
        unique_variants = []
        for t in variant_types:
            if t not in unique_variants:
                unique_variants.append(t)

        return typing.Union[tuple(variant_types)]

    if literal.enum_type is not None:
        # We can hold the variants a proxy class that is also type 'Enum', s.t.
        # we can parse the variants and define the object in the param map
        # code.

        class _VariantCarrier(enum.Enum): ...

        _VariantCarrier._variants = literal.enum_type.values
        # Use param name to uniquely identify each enum
        _VariantCarrier._name = param_name
        return _VariantCarrier

    raise NotImplementedError(
        f"The flyte literal {literal} cannot be transformed to a python type."
    )


def _best_effort_default_val(t: typing.T):
    """Produce a "best-effort" default value given a python type."""

    if t in _primitive_table:
        return _primitive_table[t]

    if t is list:
        return []

    file_like_table = {
        LatchDir: LatchDir("latch:///foobar"),
        LatchFile: LatchFile("latch:///foobar"),
    }
    if t in file_like_table:
        return file_like_table[t]

    if type(t) is enum.EnumMeta:
        return f"{t._name}.{t._variants[0]}"

    if get_origin(t) is None:
        raise NotImplementedError(
            f"Unable to produce a best-effort value for the python type {t}"
        )

    if get_origin(t) is list:
        list_args = get_args(t)
        if len(list_args) == 0:
            return []
        return [_best_effort_default_val(arg) for arg in list_args]

    if get_origin(t) is typing.Union:
        return _best_effort_default_val(get_args(t)[0])

    raise NotImplementedError(
        f"Unable to produce a best-effort value for the python type {t}"
    )



"""
DEPRECATED

Mimics channel operators from Nextflow, using the correspondence Channel --> Python Dictionary
"""

import re
from itertools import product
from typing import Any, Callable, Dict, List, Optional, Tuple, Union


def _combine(item1: Any, item2: Any):
    """
    Combines two items for use in *_join functions. The rules followed are:

    - If both items are lists, the lists are concatenated
    - If one of the items is a list, the other is appended to that list
    - Otherwise, the output is a new list containing both items

    This is so that composition of joins works as expected. We also use list
    addition so as to not modify the input items and instead return a new copy.
    """

    if isinstance(item1, List) and isinstance(item2, List):
        return item1 + item2
    elif isinstance(item1, List):
        return item1 + [item2]
    elif isinstance(item2, List):
        return [item1] + item2
    else:
        return [item1, item2]


def left_join(left: Dict[str, Any], right: Dict[str, Any]) -> Dict[str, Any]:
    """A standard left join of two dictionaries, joining on their keys"""
    output = {}
    for key in left:
        if key in right:
            output[key] = _combine(left[key], right[key])
        else:
            output[key] = left[key]
    return output


def right_join(left: Dict[str, Any], right: Dict[str, Any]) -> Dict[str, Any]:
    """A standard right join of two dictionaries, joining on their keys"""
    output = {}
    for key in right:
        if key in left:
            output[key] = _combine(left[key], right[key])
        else:
            output[key] = right[key]
    return output


def inner_join(left: Dict[str, Any], right: Dict[str, Any]) -> Dict[str, Any]:
    """A standard inner join of two dictionaries, joining on their keys"""
    output = {}
    for key in left:
        if key in right:
            output[key] = _combine(left[key], right[key])
    return output


def outer_join(left: Dict[str, Any], right: Dict[str, Any]) -> Dict[str, Any]:
    """A standard outer join of two dictionaries, joining on their keys"""
    output = {}
    for key in left:
        if key in right:
            output[key] = _combine(left[key], right[key])
        else:
            output[key] = left[key]
    for key in right:
        if key not in output:
            output[key] = right[key]
    return output


def group_tuple(channel: List[Tuple], key_index: Optional[int] = None) -> List[Tuple]:
    """
    Operator to mimic the `groupTuple` construct from Nextflow:

        The `groupTuple` operator collects tuples (or lists) of values emitted
        by the source channel grouping together the elements that share the same
        key. Finally it emits a new tuple object for each distinct key collected.

    Args:
        channel: A list of tuples to be grouped by key_index
        key_index: Which index of the tuple to match against - if not provided,
            defaults to 0

    Example:
        >>> channel = [(1,'A'), (1,'B'), (2,'C'), (3, 'B'), (1,'C'), (2, 'A'), (3, 'D')]
        >>> group_tuple(channel) # key_index defaults to grouping by the first element (index 0)
        [(1, ['A', 'B', 'C']), (2, ['C', 'A']), (3, 'B', 'D')]
    """

    output = {}
    if key_index is None:
        key_index = 0
    for element in channel:
        if not (0 <= key_index < len(element)):
            raise ValueError(f"Key Index {key_index} too large for element {element}")
        key = element[key_index]
        if key not in output:
            output[key] = tuple(
                [[k] if i != key_index else k for i, k in enumerate(element)]
            )
        else:
            for i, k in enumerate(element):
                if i == key_index:
                    continue
                output[key][i].append(k)
    return list(output.values())


def latch_filter(
    channel: List[Any],
    predicate: Union[Callable, re.Pattern, type, None],
) -> List[Any]:
    """Filters a given list with either a predicate, a regex, or a type"""
    if isinstance(predicate, Callable):
        return list(filter(predicate, channel))
    elif isinstance(predicate, re.Pattern):

        def filter_func(list_item: Any):
            if isinstance(list_item, str):
                return predicate.match(list_item)
            return False

        return list(filter(filter_func, channel))
    elif isinstance(predicate, type):

        def filter_func(list_item: Any):
            return isinstance(list_item, predicate)

        return list(filter(filter_func, channel))
    else:
        return channel


def combine(
    channel_0: List[Any],
    channel_1: List[Any],
    by: Optional[int] = None,
) -> Union[List, Dict[str, List[Any]]]:
    """
    Creates a Cartesian product of the two provided channels, with the option to first group
    the elements of the channels by a certain index and then doing individual products on the
    groups.

    Args:
        channel_0: A list. If by is provided, all elements must be tuples of the same length
        channel_1: A list. If by is provided, all elements must be tuples of the same length
            as channel_0
        by: If provided, which index to group by first.

    Example:
        >>> c0 = ['hello', 'ciao']
        >>> c1 = [1, 2, 3]
        >>> combine(c0, c1)
        [('hello', 1), ('hello', 2), ('hello', 3), ('ciao', 1), ('ciao', 2), ('ciao', 3)]
    """
    if by is not None:
        output = {}
        for element in channel_0:
            if not isinstance(element, tuple):
                raise ValueError(f"`by` is provided, but {element} is not a tuple.")
            if not (0 <= by < len(element)):
                raise ValueError(f"Combine index {by} too large for element {element}")
            if element[by] not in output:
                output[element[by]] = [[], []]
            output[element[by]][0].append(element[:by] + element[by + 1 :])
        for element in channel_1:
            if not isinstance(element, tuple):
                raise ValueError(f"`by` is provided, but {element} is not a tuple.")
            if not (0 <= by < len(element)):
                raise ValueError(f"Combine index {by} too large for element {element}")
            if element[by] not in output:
                output[element[by]] = [[], []]
            output[element[by]][1].append(element[:by] + element[by + 1 :])
        final_output = []
        for key in output:
            prod = list(product(*output[key]))
            for p1, p2 in prod:
                final_output.append((key,) + p1 + p2)
        return final_output
    return list(product(channel_0, channel_1))



from wf.task import task

from latch.resources.workflow import workflow
from latch.types.directory import LatchOutputDir
from latch.types.file import LatchFile
from latch.types.metadata import LatchAuthor, LatchMetadata, LatchParameter

metadata = LatchMetadata(
    display_name="WF_NAME",
    author=LatchAuthor(
        name="AUTHOR_NAME",
    ),
    parameters={
        "input_file": LatchParameter(
            display_name="Input File",
            batch_table_column=True,  # Show this parameter in batched mode.
        ),
        "output_directory": LatchParameter(
            display_name="Output Directory",
            batch_table_column=True,  # Show this parameter in batched mode.
        ),
    },
)


@workflow(metadata)
def template_workflow(
    input_file: LatchFile, output_directory: LatchOutputDir
) -> LatchFile:
    return task(input_file=input_file, output_directory=output_directory)



import re
from dataclasses import Field, asdict, dataclass, field
from enum import Enum
from pathlib import Path
from textwrap import indent
from typing import Any, ClassVar, Dict, List, Optional, Protocol, Tuple, Type, Union

import click
import yaml
from typing_extensions import TypeAlias

from latch_cli.snakemake.config.utils import validate_snakemake_type
from latch_cli.utils import identifier_suffix_from_str

from .directory import LatchDir
from .file import LatchFile


@dataclass
class LatchRule:
    """Class describing a rule that a parameter input must follow"""

    regex: str
    """A string regular expression which inputs must match"""
    message: str
    """The message to render when an input does not match the regex"""

    @property
    def dict(self):
        return asdict(self)

    def __post_init__(self):
        try:
            re.compile(self.regex)
        except re.error as e:
            raise ValueError(f"Malformed regex {self.regex}: {e.msg}")


class LatchAppearanceEnum(Enum):
    line = "line"
    paragraph = "paragraph"


@dataclass(frozen=True)
class MultiselectOption:
    name: str
    value: object


@dataclass(frozen=True)
class Multiselect:
    options: List[MultiselectOption] = field(default_factory=list)
    allow_custom: bool = False


# backwards compatibility
LatchAppearanceType = LatchAppearanceEnum

LatchAppearance: TypeAlias = Union[LatchAppearanceEnum, Multiselect]


@dataclass
class LatchAuthor:
    """Class describing metadata about the workflow author"""

    name: Optional[str] = None
    """The name of the author"""
    email: Optional[str] = None
    """The email of the author"""
    github: Optional[str] = None
    """A link to the github profile of the author"""


@dataclass(frozen=True)
class FlowBase:
    """Parent class for all flow elements

    Available flow elements:

    * :class:`~latch.types.metadata.Params`

    * :class:`~latch.types.metadata.Text`

    * :class:`~latch.types.metadata.Title`

    * :class:`~latch.types.metadata.Section`

    * :class:`~latch.types.metadata.Spoiler`

    * :class:`~latch.types.metadata.Fork`
    """

    ...


@dataclass(frozen=True, init=False)
class Section(FlowBase):
    """Flow element that displays a child flow in a card with a given title

    Example:


    .. image:: ../assets/flow-example/flow_example_1.png
        :alt: Example of a user interface for a workflow with a custom flow

    .. image:: ../assets/flow-example/flow_example_spoiler.png
        :alt: Example of a spoiler flow element


    The `LatchMetadata` for the example above can be defined as follows:

    .. code-block:: python

        from latch.types import LatchMetadata, LatchParameter
        from latch.types.metadata import FlowBase, Section, Text, Params, Fork, Spoiler
        from latch import workflow

        flow = [
            Section(
                "Samples",
                Text(
                    "Sample provided has to include an identifier for the sample (Sample name)"
                    " and one or two files corresponding to the reads (single-end or paired-end, respectively)"
                ),
                Fork(
                    "sample_fork",
                    "Choose read type",
                    paired_end=ForkBranch("Paired-end", Params("paired_end")),
                    single_end=ForkBranch("Single-end", Params("single_end")),
                ),
            ),
            Section(
                "Quality threshold",
                Text(
                    "Select the quality value in which a base is qualified."
                    "Quality value refers to a Phred quality score"
                ),
                Params("quality_threshold"),
            ),
            Spoiler(
                "Output directory",
                Text("Name of the output directory to send results to."),
                Params("output_directory"),
            ),
        ]

        metadata = LatchMetadata(
            display_name="fastp - Flow Tutorial",
            author=LatchAuthor(
                name="LatchBio",
            ),
            parameters={
                "sample_fork": LatchParameter(),
                "paired_end": LatchParameter(
                    display_name="Paired-end reads",
                    description="FASTQ files",
                    batch_table_column=True,
                ),
                "single_end": LatchParameter(
                    display_name="Single-end reads",
                    description="FASTQ files",
                    batch_table_column=True,
                ),
                "output_directory": LatchParameter(
                    display_name="Output directory",
                ),
            },
            flow=flow,
        )

        @workflow(metadata)
        def fastp(
            sample_fork: str,
            paired_end: PairedEnd,
            single_end: Optional[SingleEnd] = None,
            output_directory: str = "fastp_results",
        ) -> LatchDir:
            ...
    """

    section: str
    """Title of the section"""
    flow: List[FlowBase]
    """Flow displayed in the section card"""

    def __init__(self, section: str, *flow: FlowBase):
        object.__setattr__(self, "section", section)
        object.__setattr__(self, "flow", list(flow))


@dataclass(frozen=True)
class Text(FlowBase):
    """Flow element that displays a markdown string"""

    text: str
    """Markdown body text"""


@dataclass(frozen=True)
class Title(FlowBase):
    """Flow element that displays a markdown title"""

    title: str
    """Markdown title text"""


@dataclass(frozen=True, init=False)
class Params(FlowBase):
    """Flow element that displays parameter widgets"""

    params: List[str]
    """
    Names of parameters whose widgets will be displayed.
    Order is preserved. Duplicates are allowed
    """

    def __init__(self, *args: str):
        object.__setattr__(self, "params", list(args))


@dataclass(frozen=True, init=False)
class Spoiler(FlowBase):
    """Flow element that displays a collapsible card with a given title"""

    spoiler: str
    """Title of the spoiler"""
    flow: List[FlowBase]
    """Flow displayed in the spoiler card"""

    def __init__(self, spoiler: str, *flow: FlowBase):
        object.__setattr__(self, "spoiler", spoiler)
        object.__setattr__(self, "flow", list(flow))


@dataclass(frozen=True, init=False)
class ForkBranch:
    """Definition of a :class:`~latch.types.metadata.Fork` branch"""

    display_name: str
    """String displayed in the fork's multibutton"""
    flow: List[FlowBase]
    """Child flow displayed in the fork card when the branch is active"""

    def __init__(self, display_name: str, *flow: FlowBase):
        object.__setattr__(self, "display_name", display_name)
        object.__setattr__(self, "flow", list(flow))


@dataclass(frozen=True, init=False)
class Fork(FlowBase):
    """Flow element that displays a set of mutually exclusive alternatives

    Displays a title, followed by a horizontal multibutton for selecting a branch,
    then a card for the active branch
    """

    fork: str
    """Name of a `str`-typed parameter to store the active branch's key"""
    display_name: str
    """Title shown above the fork selector"""
    flows: Dict[str, ForkBranch]
    """
    Mapping between branch keys to branch definitions.
    Order determines the order of options in the multibutton
    """

    def __init__(self, fork: str, display_name: str, **flows: ForkBranch):
        object.__setattr__(self, "fork", fork)
        object.__setattr__(self, "display_name", display_name)
        object.__setattr__(self, "flows", flows)


@dataclass
class LatchParameter:
    """Class for organizing parameter metadata"""

    display_name: Optional[str] = None
    """The name used to display the parameter on Latch Console"""
    description: Optional[str] = None
    """The description of the parameter's role in the workflow"""
    hidden: bool = False
    """Whether or not the parameter should be hidden by default"""
    section_title: Optional[str] = None
    """Whether this parameter should start a new section"""
    placeholder: Optional[str] = None
    """
    What should be rendered as a placeholder in the input box
    of the parameter before any value is inputed.
    """
    comment: Optional[str] = None
    """Any comment on the parameter itself"""
    output: bool = False
    """
    Whether or not this parameter is an output (used to disable
    path validation before launching a workflow)
    """
    batch_table_column: bool = False
    """
    Whether this parameter should be given a column in the batch
    table at the top of the workflow inputs
    """
    allow_dir: bool = True
    """
    Whether or not this parameter should accept directories in UI
    """
    allow_file: bool = True
    """
    Whether or not this parameter should accept files in UI.
    """
    appearance_type: LatchAppearance = LatchAppearanceEnum.line
    """
    Whether the parameter should be rendered as a line or paragraph
    (must be exactly one of either LatchAppearanceType.line or
    LatchAppearanceType.paragraph)
    """
    rules: List[LatchRule] = field(default_factory=list)
    """
    A list of LatchRule objects that inputs to this parameter must follow
    """
    detail: Optional[str] = None
    samplesheet: Optional[bool] = None
    """
    Use samplesheet input UI. Allows importing from Latch Registry.
    Parameter type must be a list of dataclasses
    """
    _custom_ingestion: Optional[str] = None

    def __str__(self):
        metadata_yaml = yaml.safe_dump(self.dict, sort_keys=False)
        if self.description is not None:
            return f"{self.description}\n{metadata_yaml}"
        return metadata_yaml

    @property
    def dict(self):
        parameter_dict: Dict[str, Any] = {"display_name": self.display_name}

        if self.output:
            parameter_dict["output"] = True
        if self.batch_table_column:
            parameter_dict["batch_table_column"] = True
        if self.samplesheet:
            parameter_dict["samplesheet"] = True

        temp_dict: Dict[str, Any] = {"hidden": self.hidden}
        if self.section_title is not None:
            temp_dict["section_title"] = self.section_title
        if self._custom_ingestion is not None:
            temp_dict["custom_ingestion"] = self._custom_ingestion

        parameter_dict["_tmp"] = temp_dict

        appearance_dict: Dict[str, Any]
        if isinstance(self.appearance_type, LatchAppearanceEnum):
            appearance_dict = {"type": self.appearance_type.value}
        elif isinstance(self.appearance_type, Multiselect):
            appearance_dict = {"multiselect": asdict(self.appearance_type)}
        else:
            appearance_dict = {}

        if self.placeholder is not None:
            appearance_dict["placeholder"] = self.placeholder
        if self.comment is not None:
            appearance_dict["comment"] = self.comment
        if self.detail is not None:
            appearance_dict["detail"] = self.detail

        appearance_dict["file_type"] = (
            "ANY"
            if self.allow_file and self.allow_dir
            else "FILE" if self.allow_file else "DIR" if self.allow_dir else "NONE"
        )

        parameter_dict["appearance"] = appearance_dict

        if len(self.rules) > 0:
            rules = []
            for rule in self.rules:
                rules.append(rule.dict)
            parameter_dict["rules"] = rules

        return {"__metadata__": parameter_dict}


# https://stackoverflow.com/questions/54668000/type-hint-for-an-instance-of-a-non-specific-dataclass
class _IsDataclass(Protocol):
    __dataclass_fields__: ClassVar[Dict[str, Field]]


ParameterType: TypeAlias = Union[
    Type[None],
    Type[int],
    Type[float],
    Type[str],
    Type[bool],
    Type[Enum],
    Type[_IsDataclass],
    Type[List["ParameterType"]],
    Type[LatchFile],
    Type[LatchDir],
]


@dataclass
class SnakemakeParameter(LatchParameter):
    type: Optional[ParameterType] = None
    """
    The python type of the parameter.
    """
    # todo(ayush): needs to be typed properly
    default: Optional[Any] = None


@dataclass
class SnakemakeFileParameter(SnakemakeParameter):
    """
    Deprecated: use `file_metadata` keyword in `SnakemakeMetadata` instead
    """

    type: Optional[
        Union[
            Type[LatchFile],
            Type[LatchDir],
        ]
    ] = None
    """
    The python type of the parameter.
    """
    path: Optional[Path] = None
    """
    The path where the file passed to this parameter will be copied.
    """
    config: bool = False
    """
    Whether or not the file path is exposed in the Snakemake config
    """
    download: bool = False
    """
    Whether or not the file is downloaded in the JIT step
    """


@dataclass
class SnakemakeFileMetadata:
    path: Path
    """
    The local path where the file passed to this parameter will be copied
    """
    config: bool = False
    """
    If `True`, expose the file in the Snakemake config
    """
    download: bool = False
    """
    If `True`, download the file in the JIT step
    """


@dataclass
class LatchMetadata:
    """Class for organizing workflow metadata

    Example:

    .. code-block:: python

        from latch.types import LatchMetadata, LatchAuthor, LatchRule, LatchAppearanceType

        metadata = LatchMetadata(
            parameters={
                "read1": LatchParameter(
                    display_name="Read 1",
                    description="Paired-end read 1 file to be assembled.",
                    hidden=True,
                    section_title="Sample Reads",
                    placeholder="Select a file",
                    comment="This is a comment",
                    output=False,
                    appearance_type=LatchAppearanceType.paragraph,
                    rules=[
                        LatchRule(
                            regex="(.fasta|.fa|.faa|.fas)$",
                            message="Only .fasta, .fa, .fas, or .faa extensions are valid"
                        )
                    ],
                    batch_table_column=True,  # Show this parameter in batched mode.
                    # The below parameters will be displayed on the side bar of the workflow
                    documentation="https://github.com/author/my_workflow/README.md",
                    author=LatchAuthor(
                        name="Workflow Author",
                        email="licensing@company.com",
                        github="https://github.com/author",
                    ),
                    repository="https://github.com/author/my_workflow",
                    license="MIT",
                    # If the workflow is public, display it under the defined categories on Latch to be more easily discovered by users
                    tags=["NGS", "MAG"],
                ),
        )

        @workflow(metadata)
        def wf(read1: LatchFile):
            ...

    """

    display_name: str
    """The human-readable name of the workflow"""
    author: LatchAuthor
    """ A `LatchAuthor` object that describes the author of the workflow"""
    documentation: Optional[str] = None
    """A link to documentation for the workflow itself"""
    repository: Optional[str] = None
    """A link to the repository where the code for the workflow is hosted"""
    license: str = "MIT"
    """A SPDX identifier"""
    parameters: Dict[str, LatchParameter] = field(default_factory=dict)
    """A dictionary mapping parameter names (strings) to `LatchParameter` objects"""
    wiki_url: Optional[str] = None
    video_tutorial: Optional[str] = None
    tags: List[str] = field(default_factory=list)
    flow: List[FlowBase] = field(default_factory=list)

    no_standard_bulk_execution: bool = False
    """
    Disable the standard CSV-based bulk execution. Intended for workflows that
    support an alternative way of processing bulk data e.g. using a samplesheet
    parameter
    """
    _non_standard: Dict[str, object] = field(default_factory=dict)

    @property
    def dict(self):
        metadata_dict = asdict(self)
        # remove parameters since that will be handled by each parameters' dict() method
        del metadata_dict["parameters"]
        metadata_dict["license"] = {"id": self.license}

        # flows override all other rendering, so disable them entirely if not provided
        if len(self.flow) == 0:
            del metadata_dict["flow"]

        for key in self._non_standard:
            metadata_dict[key] = self._non_standard[key]

        return {"__metadata__": metadata_dict}

    def __str__(self):
        def _parameter_str(t: Tuple[str, LatchParameter]):
            parameter_name, parameter_meta = t
            return f"{parameter_name}:\n" + indent(
                str(parameter_meta), "  ", lambda _: True
            )

        metadata_yaml = yaml.safe_dump(self.dict, sort_keys=False)
        parameter_yaml = "".join(map(_parameter_str, self.parameters.items()))
        return (
            metadata_yaml + "Args:\n" + indent(parameter_yaml, "  ", lambda _: True)
        ).strip("\n ")


@dataclass
class DockerMetadata:
    """Class describing credentials for private docker repositories"""

    username: str
    """
    The account username for the private repository
    """
    secret_name: str
    """
    The name of the Latch Secret that contains the password for the private repository
    """


@dataclass
class EnvironmentConfig:
    """Class describing environment for spawning Snakemake tasks"""

    use_conda: bool = False
    """
    Use Snakemake `conda` directive to spawn tasks in conda environments
    """
    use_container: bool = False
    """
    Use Snakemake `container` directive to spawn tasks in Docker containers
    """
    container_args: List[str] = field(default_factory=list)
    """
    Additional arguments to use when running Docker containers
    """


FileMetadata: TypeAlias = Dict[str, Union[SnakemakeFileMetadata, "FileMetadata"]]


@dataclass
class SnakemakeMetadata(LatchMetadata):
    """Class for organizing Snakemake workflow metadata"""

    output_dir: Optional[LatchDir] = None
    """
    Directory for snakemake workflow outputs
    """
    name: Optional[str] = None
    """
    Name of the workflow
    """
    docker_metadata: Optional[DockerMetadata] = None
    """
    Credentials configuration for private docker repositories
    """
    env_config: EnvironmentConfig = field(default_factory=EnvironmentConfig)
    """
    Environment configuration for spawning Snakemake tasks
    """
    parameters: Dict[str, SnakemakeParameter] = field(default_factory=dict)
    """
    A dictionary mapping parameter names (strings) to `SnakemakeParameter` objects
    """
    file_metadata: FileMetadata = field(default_factory=dict)
    """
    A dictionary mapping parameter names to `SnakemakeFileMetadata` objects
    """
    cores: int = 4
    """
    Number of cores to use for Snakemake tasks (equivalent of Snakemake's `--cores` flag)
    """

    def validate(self):

        for name, param in self.parameters.items():
            if param.default is None:
                continue
            try:
                validate_snakemake_type(name, param.type, param.default)
            except ValueError as e:
                click.secho(e, fg="red")
                raise click.exceptions.Exit(1)

    def __post_init__(self):
        self.validate()

        if self.name is None:
            self.name = (
                f"snakemake_{identifier_suffix_from_str(self.display_name.lower())}"
            )

        global _snakemake_metadata
        _snakemake_metadata = self


_snakemake_metadata: Optional[SnakemakeMetadata] = None



import re

MAX_LINES = 10

# for parsing out ansi escape codes during register for pretty printing so
# that e.g. carriage returns (\r) don't break the printer
ANSI_REGEX = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")



import os

import click
import gql
from latch_sdk_gql.execute import execute


def rename_current_execution(name: str):
    """Rename the current execution.

    Useful for naming executions based on inputs. If this function is called
    outside of an execution context, it is a noop.
    """
    token = os.environ.get("FLYTE_INTERNAL_EXECUTION_ID", None)
    if token is None:
        # noop during local execution / testing
        click.secho(
            "Running in local execution context - skipping rename.",
            dim=True,
            italic=True,
        )
        return

    execute(
        gql.gql("""
            mutation RenameExecution($argName: String!, $argToken: String!) {
                renameExecutionByToken(input: {argToken: $argToken, argName: $argName}) {
                    clientMutationId
                }
            }
        """),
        {"argName": name, "argToken": token},
    )



from textwrap import dedent

import gql
from gql.transport.exceptions import TransportQueryError

from latch.ldata.type import LatchPathError, LDataNodeType

from .node import get_node_data
from .utils import query_with_retry


def remote_copy(src: str, dst: str, create_parents: bool = False) -> None:
    node_data = get_node_data(src, dst, allow_resolve_to_parent=True)

    src_data = node_data.data[src]
    dst_data = node_data.data[dst]
    acc_id = node_data.acc_id

    path_by_id = {v.id: k for k, v in node_data.data.items()}

    if not src_data.exists():
        raise LatchPathError("not found", src, acc_id)

    if dst_data.exists() and dst_data.type in {LDataNodeType.obj, LDataNodeType.link}:
        raise LatchPathError("object already exists at path", dst, acc_id)

    # if the destination exists and is a directory, use the source name
    path = None
    if not dst_data.exists():
        if not dst_data.is_direct_parent() and not create_parents:
            raise LatchPathError("no such Latch file or directory", dst, acc_id)
        path = dst_data.remaining

    try:
        query_with_retry(
            gql.gql("""
            mutation Copy(
                $argSrcNode: BigInt!
                $argDstParent: BigInt!
                $argPath: String
            ) {
                ldataCopyPath(
                    input: {
                        argSrcNode: $argSrcNode
                        argDstParent: $argDstParent
                        argPath: $argPath
                    }
                ) {
                    clientMutationId
                }
            }"""),
            {
                "argSrcNode": src_data.id,
                "argDstParent": dst_data.id,
                "argPath": path,
            },
        )
    except TransportQueryError as e:
        if e.errors is None or len(e.errors) == 0:
            raise e

        msg: str = e.errors[0]["message"]

        if msg.startswith("Permission denied on node"):
            node_id = msg.rsplit(" ", 1)[1]
            path = path_by_id[node_id]

            raise LatchPathError("permission denied", path, acc_id)
        elif msg == "Refusing to make node its own parent":
            raise LatchPathError(f"is a parent of {src}", dst, acc_id)
        elif msg == "Refusing to parent node to an object node":
            raise LatchPathError(f"object exists at path", dst, acc_id)
        elif msg == "Refusing to move a share link (or into a share link)":
            raise LatchPathError(
                "is a share link",
                src if src_data.type is LDataNodeType.link else dst,
                acc_id,
            )
        elif msg.startswith("Refusing to copy account root"):
            raise LatchPathError("is an account root", src, acc_id)
        elif msg.startswith("Refusing to copy removed node"):
            raise LatchPathError("not found", src, acc_id)
        elif msg.startswith("Refusing to copy already in-transit node"):
            raise LatchPathError("copy already in progress", src, acc_id)
        elif msg == "Conflicting object in destination":
            raise LatchPathError("object exists at path", dst, acc_id)

        raise LatchPathError(str(e), src, acc_id)



from dataclasses import dataclass
from typing import Union


@dataclass
class _SyntaxError(BaseException):
    filename: str
    lineno: int
    offset: int
    text: str
    end_lineno: int
    end_offset: int


class _FlytekitError(BaseException): ...


_HandledError = Union[_SyntaxError, _FlytekitError]



from pathlib import Path
from textwrap import dedent
from typing import List

import click

from latch.ldata._transfer.download import download as _download
from latch.ldata._transfer.progress import Progress
from latch.ldata._transfer.remote_copy import remote_copy as _remote_copy
from latch.ldata._transfer.upload import upload as _upload
from latch.ldata.type import LatchPathError
from latch_cli.services.cp.glob import expand_pattern
from latch_cli.utils import human_readable_time, with_si_suffix
from latch_cli.utils.path import get_path_error, is_remote_path


def _copy_and_print(src: str, dst: str, progress: Progress) -> None:
    _remote_copy(src, dst)
    if progress != Progress.none:
        click.echo(dedent(f"""
            {click.style("Copy Requested.", fg="green")}
            {click.style("Source: ", fg="blue")}{(src)}
            {click.style("Destination: ", fg="blue")}{(dst)}"""))


def _download_and_print(src: str, dst: Path, progress: Progress, verbose: bool) -> None:
    if progress != Progress.none:
        click.secho(f"Downloading {dst.name}", fg="blue")
    res = _download(src, dst, progress, verbose)
    if progress != Progress.none:
        click.echo(dedent(f"""
			{click.style("Download Complete", fg="green")}
			{click.style("Time Elapsed: ", fg="blue")}{human_readable_time(res.total_time)}
			{click.style("Files Downloaded: ", fg="blue")}{res.num_files} ({with_si_suffix(res.total_bytes)})
			"""))


# todo(ayush): come up with a better behavior scheme than unix cp
def cp(
    srcs: List[str],
    dest: str,
    *,
    progress: Progress,
    verbose: bool,
    expand_globs: bool,
):
    dest_remote = is_remote_path(dest)

    for src in srcs:
        src_remote = is_remote_path(src)

        try:
            if src_remote and not dest_remote:
                if expand_globs:
                    [
                        _download_and_print(p, Path(dest), progress, verbose)
                        for p in expand_pattern(src)
                    ]
                else:
                    _download_and_print(src, Path(dest), progress, verbose)
            elif not src_remote and dest_remote:
                if progress != Progress.none:
                    click.secho(f"Uploading {src}", fg="blue")
                res = _upload(src, dest, progress=progress, verbose=verbose)
                if progress != Progress.none:
                    click.echo(dedent(f"""
                        {click.style("Upload Complete", fg="green")}
                        {click.style("Time Elapsed: ", fg="blue")}{human_readable_time(res.total_time)}
                        {click.style("Files Uploaded: ", fg="blue")}{res.num_files} ({with_si_suffix(res.total_bytes)})
                        """))
            elif src_remote and dest_remote:
                if expand_globs:
                    [_copy_and_print(p, dest, progress) for p in expand_pattern(src)]
                else:
                    _copy_and_print(src, dest, progress)
            else:
                raise ValueError(
                    dedent(f"""
                    `latch cp` cannot be used for purely local file copying.

                    Please ensure at least one of your arguments is a remote path (beginning with `latch://`)
                    """).strip("\n"),
                )
        except LatchPathError as e:
            click.secho(get_path_error(e.remote_path, e.message, e.acc_id), fg="red")
            raise click.exceptions.Exit(1) from e
        except Exception as e:
            click.secho(str(e), fg="red")
            raise click.exceptions.Exit(1) from e



"""
Minimal template workflow to show how to run workflows that use nextflow in Latch

For a more comprehensive template, see the assemble_and_sort workflow
For examples on how to use nextflow in Latch, see https://docs.latch.bio/examples/workflows_examples.html
"""
from wf.task import run_nfcore_fetchngs

from latch.resources.launch_plan import LaunchPlan
from latch.resources.workflow import workflow
from latch.types.directory import LatchDir, LatchOutputDir
from latch.types.file import LatchFile
from latch.types.metadata import LatchAuthor, LatchMetadata, LatchParameter, LatchRule

"""The metadata included here will be injected into your interface."""
metadata = LatchMetadata(
    display_name="nf-core/fetchngs",
    documentation="https://nf-co.re/fetchngs/1.8/usage",
    author=LatchAuthor(
        name="nf-core",
    ),
    parameters={
        "ids": LatchParameter(
            display_name="IDs CSV",
            description="CSV file containing ids from SRA, ENA, DDBJ or Synapse",
            batch_table_column=True,  # Show this parameter in batched mode
            rules=[
                # validate the input file using regex
                LatchRule(
                    regex="(.csv)$",
                    message="Only csv extension is valid",
                )
            ],
        ),
        "output_directory": LatchParameter(
            display_name="Output Directory",
            description="Where to place the result file.",
            batch_table_column=True,  # Show this parameter in batched mode.
        ),
    },
)


@workflow(metadata)
def nfcore_fetchngs(ids: LatchFile, output_directory: LatchOutputDir) -> LatchDir:
    return run_nfcore_fetchngs(ids=ids, output_directory=output_directory)


"""
Add test data with a LaunchPlan. Provide default values in a dictionary with
the parameter names as the keys. These default values will be available under
the 'Test Data' dropdown at console.latch.bio.
"""
LaunchPlan(
    nfcore_fetchngs,
    "Test Data",
    {
        "ids": LatchFile("s3://latch-public/init/nfcore_example_ids.csv"),
        "output_directory": LatchOutputDir("latch:///nfcore_fetchngs_output"),
    },
)



from typing import List, TypedDict

try:
    from functools import cache
except ImportError:
    from functools import lru_cache as cache

import gql
from latch_sdk_gql.execute import execute


class Child(TypedDict):
    name: str


class ChildLdataTreeEdgesNode(TypedDict):
    child: Child


class ChildLdataTreeEdges(TypedDict):
    nodes: List[ChildLdataTreeEdgesNode]


class LdataResolvePathData(TypedDict):
    childLdataTreeEdges: ChildLdataTreeEdges


@cache
def _get_immediate_children_of_node(path: str) -> List[str]:
    lrpd: LdataResolvePathData = execute(
        gql.gql("""
            query MyQuery($argPath: String!) {
                ldataResolvePathData(argPath: $argPath) {
                    childLdataTreeEdges(
                        filter: {child: {removed: {equalTo: false}}}
                    ) {
                        nodes {
                            child {
                                name
                            }
                        }
                    }
                }
            }
        """),
        {"argPath": path},
    )["ldataResolvePathData"]

    if lrpd is None:
        return []

    res: List[str] = []
    for node in lrpd["childLdataTreeEdges"]["nodes"]:
        res.append(node["child"]["name"])

    return res


class Team(TypedDict):
    accountId: str


class TeamMembersByUserIdNode(TypedDict):
    team: Team


class TeamMembersByUserId(TypedDict):
    nodes: List[TeamMembersByUserIdNode]


class TeamInfosByOwnerId(TypedDict):
    nodes: List[Team]


class UserInfoByAccountId(TypedDict):
    defaultAccount: str
    teamMembersByUserId: TeamMembersByUserId
    teamInfosByOwnerId: TeamInfosByOwnerId


class Bucket(TypedDict):
    bucketName: str


class LdataS3MountAccessProvensByGeneratedUsing(TypedDict):
    nodes: List[Bucket]


class LdataS3MountConfiguratorRolesByAccountIdNode(TypedDict):
    ldataS3MountAccessProvensByGeneratedUsing: LdataS3MountAccessProvensByGeneratedUsing


class LdataS3MountConfiguratorRolesByAccountId(TypedDict):
    nodes: List[LdataS3MountConfiguratorRolesByAccountIdNode]


class AccountInfoCurrent(TypedDict):
    userInfoByAccountId: UserInfoByAccountId
    ldataS3MountConfiguratorRolesByAccountId: LdataS3MountConfiguratorRolesByAccountId


@cache
def _get_known_domains_for_account() -> List[str]:
    aic: AccountInfoCurrent = execute(gql.gql("""
        query DomainCompletionQuery {
            accountInfoCurrent {
                userInfoByAccountId {
                    defaultAccount
                    teamMembersByUserId(
                        filter: { team: { account: { removed: { equalTo: false } } } }
                    ) {
                        nodes {
                            team {
                                accountId
                            }
                        }
                    }
                    teamInfosByOwnerId(filter: { account: { removed: { equalTo: false } } }) {
                        nodes {
                            accountId
                        }
                    }
                }
                ldataS3MountConfiguratorRolesByAccountId {
                    nodes {
                        ldataS3MountAccessProvensByGeneratedUsing {
                            nodes {
                                bucketName
                            }
                        }
                    }
                }
            }
        }
    """))["accountInfoCurrent"]

    ui = aic["userInfoByAccountId"]

    res: List[str] = [""]  # "" is for latch:///

    accs: List[int] = [int(ui["defaultAccount"])]
    accs.extend(
        int(tm["team"]["accountId"]) for tm in ui["teamMembersByUserId"]["nodes"]
    )
    accs.extend(int(ti["accountId"]) for ti in ui["teamInfosByOwnerId"]["nodes"])
    accs.sort()
    for x in accs:
        res.append(f"{x}.account")
        res.append(f"shared.{x}.account")

    buckets = [
        map["bucketName"]
        for role in aic["ldataS3MountConfiguratorRolesByAccountId"]["nodes"]
        for map in role["ldataS3MountAccessProvensByGeneratedUsing"]["nodes"]
    ]
    buckets.sort()
    res.extend(f"{x}.mount" for x in buckets)

    return res



from typing import Dict, List, TypedDict

import click
from latch_sdk_config.latch import config
from latch_sdk_config.user import user_config

from latch_cli.menus import SelectOption, select_tui
from latch_cli.tinyrequests import post
from latch_cli.utils import current_workspace, retrieve_or_login


def _get_workspaces() -> Dict[str, str]:
    headers = {"Authorization": f"Bearer {retrieve_or_login()}"}

    resp = post(
        url=config.api.user.list_workspaces,
        headers=headers,
    )
    resp.raise_for_status()

    data = resp.json()
    return data


class WSInfo(TypedDict):
    workspace_id: str
    name: str


def workspace():
    """Opens a terminal user interface in which a user can select the workspace
    the want to switch to.

    Like `get_executions`, this function should only be called from the CLI.
    """
    data = _get_workspaces()

    old_id = current_workspace()

    selected_marker = "\x1b[3m\x1b[2m (currently selected) \x1b[22m\x1b[23m"

    options: List[SelectOption[WSInfo]] = []
    for id, name in sorted(
        data.items(), key=lambda x: "" if x[1] == "Personal Workspace" else x[0]
    ):
        display_name = name
        if id == old_id:
            display_name = f"{name}{selected_marker}"

        options.append(
            {
                "display_name": display_name,
                "value": {
                    "workspace_id": id,
                    "name": name,
                },
            }
        )

    selected_option = select_tui(
        title="Select Workspace",
        options=options,
        clear_terminal=False,
    )

    if selected_option is None:
        return

    user_config.update_workspace(**selected_option)

    if old_id != selected_option["workspace_id"]:
        click.secho(
            f"Successfully switched to context {selected_option['name']}", fg="green"
        )
    else:
        click.secho(f"Already in context {selected_option['name']}.", fg="green")



# # variables
# # - source local/remote
# # - dest local/remote
# # - source type
# # - dest type
# # - source exists
# # - dest exists
# # - source trailing slash
# # - if source/dest remote:
# #   - domain type

# import urllib.parse
# from dataclasses import dataclass
# from enum import Enum
# from typing import Optional

# acc_id = "6612"
# bucket = "latch-sdk-test-bucket"
# node_id = "22463636"


# class Domain(Enum, str):
#     infer = ""
#     account = f"{acc_id}.account"
#     shared = "shared"
#     mount = f"{bucket}.mount"
#     shared_account = f"shared.{acc_id}.account"
#     node = f"{node_id}.node"


# @dataclass(frozen=True)
# class TestCase:
#     dir: bool
#     exists: bool
#     trailing_slash: bool


# @dataclass(frozen=True)
# class RemoteTestCase(TestCase):
#     domain: Domain


# urllib.parse.uses_netloc.append("latch")
# urllib.parse.uses_relative.append("latch")


# def generate_test_case(case: TestCase) -> str:
#     scheme: str = ""
#     netloc: str = ""
#     if isinstance(case, RemoteTestCase):
#         scheme = "latch"
#         netloc = case.domain.value


# wip: todo(ayush)



"""Service to list files in a remote directory."""

from dataclasses import dataclass
from datetime import datetime
from textwrap import dedent
from typing import List, Optional, TypedDict

import click
import dateutil.parser as dp
import gql
from latch_sdk_gql.execute import execute

from latch.ldata.type import LDataNodeType
from latch_cli.click_utils import bold
from latch_cli.utils import with_si_suffix
from latch_cli.utils.path import normalize_path


class _LdataObjectMeta(TypedDict):
    modifyTime: Optional[str]
    contentSize: Optional[int]


class _Child(TypedDict):
    name: str
    ldataObjectMeta: Optional[_LdataObjectMeta]
    type: str


class _Node(TypedDict):
    child: _Child


class _ChildLdataTreeEdges(TypedDict):
    nodes: List[_Node]


class _FinalLinkTarget(TypedDict):
    childLdataTreeEdges: _ChildLdataTreeEdges


class _LdataResolvePathData(TypedDict):
    name: str
    type: str
    ldataObjectMeta: Optional[_LdataObjectMeta]
    finalLinkTarget: _FinalLinkTarget


@dataclass(frozen=True)
class _Row:
    name: str
    type: LDataNodeType
    size: Optional[int]
    modify_time: Optional[datetime]


def ls(path: str, *, group_directories_first: bool = False):
    """Lists the children of a remote directory in Latch.

    Args:
        path: A valid remote path
        group_directories_first: Option to display directories/links before
            objects

    This function will list all of the entites under the remote directory
    specified in the path `path`. Will error if the path is invalid
    or the directory doesn't exist.

    Examples:
        >>> ls("")
            # Lists all entities in the user's root directory
        >>> ls("latch:///dir1/dir2/dir_name")
            # Lists all entities inside dir1/dir2/dir_name
    """
    if path == "":
        path = "/"

    normalized_path = normalize_path(path, assume_remote=True)

    query = execute(
        gql.gql("""
            query LdataInfo ($argPath: String!) {
                accountInfoCurrent {
                    id
                }
                ldataResolvePathData(argPath: $argPath) {
                    name
                    ldataObjectMeta {
                        modifyTime
                        contentSize
                    }
                    type
                    finalLinkTarget {
                        childLdataTreeEdges(filter: {
                            child: {
                                removed: { equalTo: false },
                                pending: { equalTo: false }
                            }
                        }) {
                            nodes {
                                child {
                                    name
                                    ldataObjectMeta {
                                        modifyTime
                                        contentSize
                                    }
                                    type
                                }
                            }
                        }
                    }
                }
            }
        """),
        {"argPath": normalized_path},
    )

    res: Optional[_LdataResolvePathData] = query["ldataResolvePathData"]
    acc_id: str = query["accountInfoCurrent"]["id"]

    if res is None:
        click.secho(
            dedent(f"""
            {bold(path)}: no such directory.

            Resolved to: {bold(normalized_path)}

            {bold("Check that:")}
            1. The target directory exists,
            2. Account {bold(acc_id)} has permissions to view the target directory, and
            3. The correct workspace is selected.

            For privacy reasons, non-viewable objects and non-existent objects are indistinguishable.
            """).strip("\n"),
            fg="red",
        )
        raise click.exceptions.Exit(1)

    nodes = res["finalLinkTarget"]["childLdataTreeEdges"]["nodes"]
    if LDataNodeType(res["type"].lower()) == LDataNodeType.obj:
        # ls object should just display the object's info
        nodes.append({"child": res})

    rows: List[_Row] = []
    for node in nodes:
        child = node["child"]

        meta = child["ldataObjectMeta"]
        size = modify_time = None
        if meta is not None:
            if meta["contentSize"] is not None:
                size = int(meta["contentSize"])
            if meta["modifyTime"] is not None:
                modify_time = dp.isoparse(meta["modifyTime"])

        rows.append(
            _Row(
                name=child["name"],
                type=LDataNodeType(child["type"].lower()),
                size=size,
                modify_time=modify_time,
            )
        )

    rows.sort(key=lambda row: row.name)
    if group_directories_first:
        rows.sort(key=lambda row: 1 if row.type == LDataNodeType.obj else 0)

    headers = [
        "  " + click.style("Size", underline=True),
        click.style("Date Modified", underline=True),
        click.style("Name", underline=True),
    ]

    click.echo(" ".join(headers))

    for row in rows:
        mt_str = f'{"-": <13}'
        size_str = f'{"-": >6}'
        if row.type == LDataNodeType.obj:
            if row.modify_time is not None:
                mt_str = click.style(
                    f'{row.modify_time.strftime("%d %b %H:%M"): <13}', fg="blue"
                )

            if row.size is not None:
                size_str = with_si_suffix(row.size, suffix="")
                size_str = click.style(f"{size_str: >6}", fg="bright_green")

        name_str = row.name
        if len(name_str) > 50:
            name_str = f"{name_str[:47]}..."

        if row.type != LDataNodeType.obj:
            color = "bright_blue"
            if row.type == LDataNodeType.link:
                color = "bright_magenta"

            name_str = click.style(f"{name_str}/", bold=True, fg=color)

        click.echo(f"{size_str} {mt_str} {name_str}")



import os
from typing import Any, Dict

import requests

NUCLEUS_URL = os.environ.get("LATCH_CLI_NUCLEUS_URL", "https://nucleus.latch.bio")
ADD_MESSAGE_ENDPOINT = f"{NUCLEUS_URL}/sdk/add-task-execution-message"


def message(typ: str, data: Dict[str, Any]) -> None:
    """Display a message prominently on the Latch console during and after a
    task execution.

    The Latch platform first processes this message internally, then displays it
    under your task's execution page.

    Args:
        typ:
            A message type that determines how your message is displayed.
            Currently one of 'info', 'warning', or 'error'.
        data:
            The data displayed on the Latch console, formatted as follows:
            ```{'title': ..., 'body': ...}```.

    Raises:
        RuntimeError: If an internal error occurs while processing the message.

    Example usage: ::

        @small_task
        def task():

            ...

            try:
                ...
            catch ValueError:
                title = 'Invalid sample ID column selected'
                body = 'Your file indicates that sample columns a, b are valid'
                message(type='error', data={'title': title, 'body': body})

            ...
    """
    task_project = os.environ.get("FLYTE_INTERNAL_TASK_PROJECT")
    task_domain = os.environ.get("FLYTE_INTERNAL_TASK_DOMAIN")
    task_name = os.environ.get("FLYTE_INTERNAL_TASK_NAME")
    task_version = os.environ.get("FLYTE_INTERNAL_TASK_VERSION")
    task_attempt_number = os.environ.get("FLYTE_ATTEMPT_NUMBER")
    execution_token = os.environ.get("FLYTE_INTERNAL_EXECUTION_ID")
    array_index = os.environ.get("FLYTE_K8S_ARRAY_INDEX")

    if task_project is None:
        print(f"Local execution message:\n[{typ}]: {data}")
        return

    response = requests.post(
        url=ADD_MESSAGE_ENDPOINT,
        json={
            "execution_token": execution_token,
            "task": {
                "project": task_project,
                "domain": task_domain,
                "name": task_name,
                "version": task_version,
            },
            "task_attempt_number": task_attempt_number,
            "task_array_index": array_index,
            "type": typ,
            "data": data,
        },
    )

    if response.status_code != 200:
        raise RuntimeError("Could not add task execution message to Latch.")
